"""Graphical User Interface (GUI) annotator module.

This module contains various annotators that define interactive ways to visualize and annotate
data loaded via a dataset parser.
"""

import abc
import collections
import json
import logging
import os

import cv2 as cv
import numpy as np

import thelper.utils

logger = logging.getLogger(__name__)


class Annotator:
    """Abstract annotation tool used to define common functions for all GUI-based I/O.

    Example configuration file::

        # ...
        "annotator": {
            # type of annotator to instantiate
            "type": "thelper.gui.ImageSegmentAnnotator",
            # ...
            # provide all extra parameters to the specialized anntator here
            "params": {
                # ...
            }
        },
        # ...

    .. seealso::
        | :class:`thelper.gui.annotators.ImageSegmentAnnotator`
        | :func:`thelper.gui.utils.create_annotator`
    """

    def __init__(self, session_name, config, save_dir, datasets):
        """Receives the annotator configuration dictionary, parses it, and sets up the basic session attributes."""
        if config is None or not isinstance(config, dict):
            raise AssertionError("invalid input config type")
        if "annotator" not in config or not config["annotator"]:
            raise AssertionError("config missing 'annotator' field")
        if save_dir is None or not os.path.isdir(save_dir):
            raise AssertionError("invalid output directory '%s'" % save_dir)
        if datasets is None or not isinstance(datasets, dict):
            raise AssertionError("invalid input dataset parser dictionary")
        self.config = config
        self.annotator_config = thelper.utils.get_key_def(["params", "parameters"], config["annotator"], {})
        self.logger = thelper.utils.get_class_logger()
        self.name = session_name
        self.save_dir = save_dir
        self.datasets = datasets
        self.logs_dir = os.path.join(self.save_dir, "logs")
        if not os.path.isdir(self.logs_dir):
            os.makedirs(self.logs_dir)
        self.annotations_dir = os.path.join(self.save_dir, "annotations")
        if not os.path.isdir(self.annotations_dir):
            os.makedirs(self.annotations_dir)
        annot_logger_path = os.path.join(self.save_dir, "logs", "annotator.log")
        annot_logger_format = logging.Formatter("[%(asctime)s - %(process)s] %(levelname)s : %(message)s")
        annot_logger_fh = logging.FileHandler(annot_logger_path)
        annot_logger_fh.setFormatter(annot_logger_format)
        self.logger.addHandler(annot_logger_fh)
        self.logger.info("created annotation log for session '%s'" % self.name)
        logstamp = thelper.utils.get_log_stamp()
        repover = thelper.__version__ + ":" + thelper.utils.get_git_stamp()
        self.logger.debug("logstamp = %s" % logstamp)
        self.logger.debug("version = %s" % repover)
        config_backup_path = os.path.join(save_dir, "logs", "config." + logstamp + ".json")
        with open(config_backup_path, "w") as fd:
            json.dump(config, fd, indent=4, sort_keys=False)

    @abc.abstractmethod
    def run(self):
        """Displays the GUI tool and blocks until it it closed by the user."""
        raise NotImplementedError


class ImageSegmentAnnotator(Annotator):
    """Annotator interface specialized for image segmentation annotation generation.

    This interface will create a GUI tool with a brush and a zoomed tooltip window that allows images
    to be painted over with predetermined class labels. The generated masks will then be saved in the
    session directory as PNG images.

    The configuration is expected to provide values for at least the following parameters:

    - ``sample_input_key``: specifies the key to use when extracting images from loaded samples. This
      is typically a string defined by the dataset parser.
    - ``labels``: provides a list of labels that will be available to use in the GUI. These labels are
      expected to be given as dictionaries that each define an ``id`` (uint8 value used in output masks),
      a ``name`` (string used for display/lookup purposes), and a color (3-element integer tuple).

    Other parameters can also be provided to alter the GUI's default behavior:

    - ``default_brush_radius``: the size of the brush at startup (default=10).
    - ``background_id``: the integer id to use for the background label (default=0).
    - ``window_zoom_crop_size``: the crop size displayed in the zoom tooltip (default=250x250).
    - ``window_zoom_size``: the size of the zoom tooltip window (default=500x500).
    - ``zoom_interp_type``: the interpolation type to use when zooming (default=cv2.INTER_NEAREST).
    - ``start_sample_idx``: the index of the first sample to display (default=0).
    - ``window_name``: the name of the main display window (default=image-segm-annotator).
    - ``window_size``: the size of the main display window (default=1000).
    - ``brush_thickness``: the size of the GUI brush tooltip border display (default=2).
    - ``gui_bar_size``: the width of the GUI bar displayed on top of the main window (default=50).
    - ``default_mask_opacity``: the default opacity of the segmentation mask (default=0.3).
    - ``default_fill_id``: the label id to fill all new masks with (default=0).

    .. seealso::
        | :class:`thelper.gui.annotators.Annotator`
    """

    # static variables used for callbacks and comms with other classes
    LATEST_PT = (-1, -1)
    LATEST_RAW_PT = (-1, -1)
    GUI_BAR_SIZE = None
    WINDOW_SIZE = None
    MOUSE_FLAGS = 0
    BRUSH_SIZE = None
    CURRENT_KEY = -1
    MASK_DIRTY = True
    GUI_DIRTY = True

    @staticmethod
    def on_press(key):
        """Callback entrypoint for pynput to register keyboard presses."""
        ImageSegmentAnnotator.CURRENT_KEY = key

    @staticmethod
    def on_mouse(event, x, y, flags, param):
        """Callback entrypoint for opencv to register mouse movement/clicks."""
        cls = ImageSegmentAnnotator
        cls.LATEST_RAW_PT = (x, y)
        cls.LATEST_PT = (x / cls.WINDOW_SIZE[0], (y - cls.GUI_BAR_SIZE) / cls.WINDOW_SIZE[1])
        cls.MOUSE_FLAGS = flags
        if event == cv.EVENT_MOUSEWHEEL:
            delta = flags >> 16
            if delta > 0:
                cls.BRUSH_SIZE = min(max(int(cls.BRUSH_SIZE * 1.1), cls.BRUSH_SIZE + 1), 100)
            else:
                cls.BRUSH_SIZE = max(min(int(cls.BRUSH_SIZE * 0.9), cls.BRUSH_SIZE - 1), 3)
        cls.GUI_DIRTY = True

    class Brush:
        """Brush manager used to refresh/draw mask contents based on mouse input."""

        def __init__(self, config):
            """Parses the input config and extracts brush-related parameters."""
            ImageSegmentAnnotator.BRUSH_SIZE = int(thelper.utils.get_key_def("default_brush_radius", config, 10))
            self.background_id = int(thelper.utils.get_key_def("background_id", config, 0))
            self.last_coords = collections.deque()

        def refresh(self, mask, label):
            """Fetches the latest mouse state and updates the mask if necessary."""
            cls = ImageSegmentAnnotator
            if not cls.GUI_DIRTY:
                return
            if 0 <= cls.LATEST_PT[0] <= 1 and 0 <= cls.LATEST_PT[1] <= 1 and \
               (cls.MOUSE_FLAGS & cv.EVENT_FLAG_LBUTTON or cls.MOUSE_FLAGS & cv.EVENT_FLAG_RBUTTON):
                coords = (int(cls.LATEST_PT[0] * mask.shape[1]), int(cls.LATEST_PT[1] * mask.shape[0]))
                self.last_coords.append(coords)
                if len(self.last_coords) == 1:
                    self.last_coords.append(coords)
                for start, stop in zip(list(self.last_coords)[0:], list(self.last_coords)[1:]):
                    if cls.MOUSE_FLAGS & cv.EVENT_FLAG_LBUTTON:
                        self.draw_stroke(mask, label["id"], start, stop)
                    else:
                        self.draw_stroke(mask, self.background_id, start, stop)
                if len(self.last_coords) > 2:
                    self.last_coords.popleft()
                cls.MASK_DIRTY = True
            else:
                self.last_coords = collections.deque()

        @staticmethod
        def draw_stroke(mask, label_id, start, end):
            """Draws a brush stroke on the mask with a given label id between two points."""
            cls = ImageSegmentAnnotator
            drag_len = cv.norm(start, end)
            brush_step_size = 1  # min(max(BRUSH_SIZE / 4, 1), 5)  # try increasing this if drawing lags
            brush_steps = int(round(max(drag_len / brush_step_size, 1)))
            coords_diff = (end[0] - start[0], end[1] - start[1])
            for step in range(brush_steps):
                alpha = step / brush_steps
                brush_coords = (int(round(start[0] + coords_diff[0] * alpha)),
                                int(round(start[1] + coords_diff[1] * alpha)))
                cv.circle(mask, brush_coords,
                          int(round((mask.shape[1] / cls.WINDOW_SIZE[0]))) * cls.BRUSH_SIZE,
                          label_id, thickness=-1, lineType=-1)

    class ZoomTooltip:
        """Zoom tooltip manager used to visualize image details based on mouse location."""

        def __init__(self, config):
            """Parses the input config and extracts zoom-related parameters."""
            self.window_zoom_crop_size = thelper.utils.str2size(thelper.utils.get_key_def("window_zoom_crop_size", config, "250x250"))
            self.window_zoom_size = thelper.utils.str2size(thelper.utils.get_key_def("window_zoom_size", config, "500x500"))
            self.window_zoom_name = "zoom"
            self.image_crop = np.empty((self.window_zoom_crop_size[1], self.window_zoom_crop_size[0], 3), dtype=np.uint8)
            self.image_zoom = np.empty((self.window_zoom_size[1], self.window_zoom_size[0], 3), dtype=np.uint8)
            self.mask_crop = np.empty((self.window_zoom_crop_size[1], self.window_zoom_crop_size[0]), dtype=np.uint8)
            self.mask_crop_color = np.empty((self.window_zoom_crop_size[1], self.window_zoom_crop_size[0], 3), dtype=np.uint8)
            self.mask_zoom = np.empty((self.window_zoom_size[1], self.window_zoom_size[0], 3), dtype=np.uint8)
            self.display_zoom = np.empty((self.window_zoom_size[1], self.window_zoom_size[0], 3), dtype=np.uint8)
            self.zoom_interp_type = thelper.utils.import_class(thelper.utils.get_key_def("zoom_interp_type", config, "cv2.INTER_NEAREST"))
            cv.namedWindow(self.window_zoom_name, cv.WINDOW_AUTOSIZE | cv.WINDOW_GUI_NORMAL)

        def refresh(self, image, mask, mask_colormap, mask_opacity, coords):
            """Fetches the latest mouse position and updates the zoom window tooltip."""
            if coords is not None:
                if self.window_zoom_crop_size[0] > image.shape[1] or self.window_zoom_crop_size[1] > image.shape[0]:
                    logger.warning("crop size too large for input image, skipping zoom refresh")
                    return
                tl = (max(coords[0] - self.window_zoom_crop_size[0] // 2, 0),
                      max(coords[1] - self.window_zoom_crop_size[1] // 2, 0))
                br = (min(tl[0] + self.window_zoom_crop_size[0], image.shape[1]),
                      min(tl[1] + self.window_zoom_crop_size[1], image.shape[0]))
                tl = (max(br[0] - self.window_zoom_crop_size[0], 0),
                      max(br[1] - self.window_zoom_crop_size[1], 0))
                if (br[0] - tl[0], br[1] - tl[1]) != self.window_zoom_crop_size:
                    logger.warning("bad zoom crop size, should never happen...?")
                else:
                    np.copyto(self.image_crop, image[tl[1]:br[1], tl[0]:br[0], ...])
                    cv.resize(self.image_crop, dsize=self.window_zoom_size, dst=self.image_zoom, interpolation=self.zoom_interp_type)
                    np.copyto(self.mask_crop, mask[tl[1]:br[1], tl[0]:br[0], ...])
                    thelper.utils.apply_color_map(self.mask_crop, mask_colormap, dst=self.mask_crop_color)
                    cv.resize(self.mask_crop_color, dsize=self.window_zoom_size, dst=self.mask_zoom, interpolation=self.zoom_interp_type)
                    cv.addWeighted(self.image_zoom, (1 - mask_opacity), self.mask_zoom, mask_opacity, 0.0, dst=self.display_zoom)
                    cv.imshow(self.window_zoom_name, self.display_zoom)

    def __init__(self, session_name, config, save_dir, datasets):
        """Parses the input samples and initializes the anntator GUI elements."""
        super().__init__(session_name, config, save_dir, datasets)
        config = self.annotator_config  # cheat for conciseness
        self.sample_input_key = thelper.utils.get_key("sample_input_key", config)
        self.logger.info("parsing datasets...")
        self.sample_count = 0
        self.sample_idx_offsets = {}
        for dataset_name, dataset in self.datasets.items():
            self.sample_idx_offsets[dataset_name] = self.sample_count
            self.sample_count += len(dataset)
            annot_dir = os.path.join(self.annotations_dir, dataset_name)
            if not os.path.isdir(annot_dir):
                os.mkdir(annot_dir)
            logstamp = thelper.utils.get_log_stamp()
            repover = thelper.__version__ + ":" + thelper.utils.get_git_stamp()
            log_content = {
                "session_name": session_name,
                "logstamp": logstamp,
                "version": repover,
                "dataset": str(dataset),
            }
            if hasattr(dataset, "samples") and isinstance(dataset.samples, list):
                log_content["samples"] = [str(sample) for sample in dataset.samples]
            dataset_log_file = os.path.join(annot_dir, "metadata.log")
            with open(dataset_log_file, "w") as fd:
                json.dump(log_content, fd, indent=4, sort_keys=False)
        self.logger.info("datasets possess a total of %d samples" % self.sample_count)
        self.curr_sample_idx = int(thelper.utils.get_key_def("start_sample_idx", config, 0))
        if self.curr_sample_idx >= self.sample_count:
            raise AssertionError("file index out-of-range (curr=%d, max=%d)" % (self.curr_sample_idx, self.sample_count - 1))
        self.window_name = thelper.utils.get_key_def("window_name", config, "image-segm-annotator")
        self.window_size = thelper.utils.get_key_def("window_size", config, 1000)
        self.brush_thickness = thelper.utils.get_key_def("brush_thickness", config, 2)
        ImageSegmentAnnotator.GUI_BAR_SIZE = thelper.utils.get_key_def("gui_bar_size", config, 50)
        self.mask_opacity = float(thelper.utils.get_key_def("default_mask_opacity", config, 0.3))
        if not (0 <= self.mask_opacity <= 1):
            raise AssertionError("unexpected opacity setting, should be in [0,1]")
        self.background_id = int(thelper.utils.get_key_def("background_id", config, 0))
        if not (0 <= self.background_id <= 255):
            raise AssertionError("background id '%s' out of 8-bit int range" % self.background_id)
        self.labels = thelper.utils.get_key("labels", config)
        self.curr_label_idx = 0
        self.label_colormap = np.zeros((256, 1, 3), dtype=np.uint8)
        for label_idx, label in enumerate(self.labels):
            if not isinstance(label, dict):
                raise AssertionError("invalid label type")
            if "id" not in label or "name" not in label or "color" not in label:
                raise AssertionError("missing some fields in label dict")
            label_id = label["id"]
            if not isinstance(label_id, int):
                raise AssertionError("unexpected label type (must be int)")
            if label_id == self.background_id or sum([label_id == lbl["id"] for lbl in self.labels]) != 1:
                raise AssertionError("duplicate label id found (%s)" % label_id)
            if not (0 <= label_id <= 255):
                raise AssertionError("label id '%s' out of 8-bit int range" % label_id)
            color = label["color"]
            if not isinstance(color, list) or len(color) != 3 or not all([isinstance(c, int) for c in color]):
                raise AssertionError("invalid label color, 3-elem integer list expected")
            self.label_colormap[label_id, 0, :] = color
        self.default_fill_id = int(thelper.utils.get_key_def("default_fill_id", config, 0))
        if self.default_fill_id not in self.labels and self.default_fill_id != self.background_id:
            raise AssertionError("unknown fill id '%s'" % self.default_fill_id)
        cv.namedWindow(self.window_name, cv.WINDOW_AUTOSIZE | cv.WINDOW_GUI_NORMAL)
        self.brush = ImageSegmentAnnotator.Brush(config)
        self.zoom = ImageSegmentAnnotator.ZoomTooltip(config)
        self.sample = None
        self.image, self.mask = self.load(self.curr_sample_idx)
        self.image_display, self.image_display_base, self.mask_display = self.refresh_layers()
        self.gui_display = None
        import pynput.keyboard
        self.listener = pynput.keyboard.Listener(on_press=ImageSegmentAnnotator.on_press)
        self.listener.start()
        cv.setMouseCallback(self.window_name, ImageSegmentAnnotator.on_mouse)

    def refresh_layers(self):
        """Updates the image, mask, and tool display layers based on the latest changes."""
        cls = ImageSegmentAnnotator
        if self.image_display_base is None:
            self.image_display_base = cv.resize(self.image, dsize=cls.WINDOW_SIZE)
            cls.MASK_DIRTY = True
        if self.mask_display is None or cls.MASK_DIRTY:
            self.mask_display = cv.resize(self.mask, dsize=cls.WINDOW_SIZE, interpolation=cv.INTER_NEAREST)
            self.mask_display = thelper.utils.apply_color_map(self.mask_display, self.label_colormap)
            cls.MASK_DIRTY = True
        if self.image_display is None or cls.MASK_DIRTY:
            self.image_display = cv.addWeighted(self.image_display_base, (1 - self.mask_opacity), self.mask_display, self.mask_opacity, 0.0)
            cls.GUI_DIRTY = True
        cls.MASK_DIRTY = False
        return self.image_display, self.image_display_base, self.mask_display

    def refresh_gui(self):
        """Updates and displays the main window based on the latest changes."""
        cls = ImageSegmentAnnotator
        if self.gui_display is None or cls.GUI_DIRTY:
            gui_shape = (self.image_display.shape[0] + cls.GUI_BAR_SIZE, self.image_display.shape[1], self.image_display.shape[2])
            if self.gui_display is None or self.gui_display.shape != gui_shape:
                self.gui_display = np.empty(gui_shape, dtype=np.uint8)
            np.copyto(self.gui_display[cls.GUI_BAR_SIZE:, ...], self.image_display)
            if cls.LATEST_PT[0] >= 0 and cls.LATEST_PT[1] >= 0:
                coords_abs = (int(cls.LATEST_PT[0] * self.image.shape[1]), int(cls.LATEST_PT[1] * self.image.shape[0]))
                self.zoom.refresh(self.image, self.mask, self.label_colormap, self.mask_opacity, coords_abs)
            curr_label = self.labels[self.curr_label_idx]
            cv.circle(self.gui_display, cls.LATEST_RAW_PT, cls.BRUSH_SIZE, curr_label["color"], self.brush_thickness)
            cv.rectangle(self.gui_display, (0, 0), (self.gui_display.shape[1], cls.GUI_BAR_SIZE), (0, 0, 0), -1)
            gui_str = "sample #%d" % self.curr_sample_idx
            # todo: update to use meta keys?
            if "path" in self.sample and isinstance(self.sample["path"], str):
                gui_str += "   path: %s" % self.sample["path"].replace('\\', '/')
            cv.putText(self.gui_display, gui_str, (10, int(cls.GUI_BAR_SIZE * 2 / 5)),
                       cv.FONT_HERSHEY_SIMPLEX, 0.40, (255, 255, 255), 1, cv.LINE_AA)
            gui_str = "current brush: %s" % curr_label["name"]
            cv.putText(self.gui_display, gui_str, (10, int(cls.GUI_BAR_SIZE * 3 / 4)),
                       cv.FONT_HERSHEY_SIMPLEX, 0.40, curr_label["color"], 1, cv.LINE_AA)
            cls.GUI_DIRTY = False
        cv.imshow(self.window_name, self.gui_display)
        cv.waitKey(1)

    def handle_keys(self):
        """Fetches the latest keyboard press and updates the annotator state accordingly."""
        import pynput.keyboard
        cls = ImageSegmentAnnotator
        nb_labels = len(self.labels)
        if cls.CURRENT_KEY != -1:
            if cls.CURRENT_KEY == pynput.keyboard.Key.down or cls.CURRENT_KEY == pynput.keyboard.KeyCode(char='s'):
                cv.imwrite(self.get_mask_path(self.curr_sample_idx), self.mask)
                self.curr_sample_idx = min(self.curr_sample_idx + 1, self.sample_count - 1)
                logger.debug("loading image+mask #%d... (max=%d)" % (self.curr_sample_idx, self.sample_count - 1))
                if self.curr_sample_idx == self.sample_count - 1:
                    logger.debug("(reached last)")
                self.image, self.mask = self.load(self.curr_sample_idx)
            elif cls.CURRENT_KEY == pynput.keyboard.Key.up or cls.CURRENT_KEY == pynput.keyboard.KeyCode(char='w'):
                cv.imwrite(self.get_mask_path(self.curr_sample_idx), self.mask)
                self.curr_sample_idx = max(self.curr_sample_idx - 1, 0)
                logger.debug("loading image+mask #%d... (max=%d)" % (self.curr_sample_idx, self.sample_count - 1))
                self.image, self.mask = self.load(self.curr_sample_idx)
            elif cls.CURRENT_KEY == pynput.keyboard.Key.right:
                self.curr_label_idx = min(self.curr_label_idx + 1, nb_labels - 1)
                curr_label = self.labels[self.curr_label_idx]
                logger.debug("swapping to label #%s : '%s'" % (curr_label["id"], curr_label["name"]))
                cls.GUI_DIRTY = True
            elif cls.CURRENT_KEY == pynput.keyboard.Key.left:
                self.curr_label_idx = max(self.curr_label_idx - 1, 0)
                curr_label = self.labels[self.curr_label_idx]
                logger.debug("swapping to label #%s : '%s'" % (curr_label["id"], curr_label["name"]))
                cls.GUI_DIRTY = True
            elif cls.CURRENT_KEY == pynput.keyboard.Key.page_up:
                self.mask_opacity = max(self.mask_opacity - 0.1, 0.0)
                logger.debug("decreasing display mask opacity to %0.1f..." % self.mask_opacity)
                self.image_display = None
            elif cls.CURRENT_KEY == pynput.keyboard.Key.page_down:
                self.mask_opacity = min(self.mask_opacity + 0.1, 1.0)
                logger.debug("increasing display mask opacity to %0.1f..." % self.mask_opacity)
                self.image_display = None
            elif cls.CURRENT_KEY == pynput.keyboard.Key.esc:
                cv.imwrite(self.get_mask_path(self.curr_sample_idx), self.mask)
                logger.debug("breaking off!")
                return True
            elif cls.CURRENT_KEY == pynput.keyboard.Key.f12:
                self.mask = np.full((self.image.shape[0], self.image.shape[1]), self.default_fill_id, dtype=np.uint8)
                self.mask_display = None
            elif cls.CURRENT_KEY == pynput.keyboard.Key.enter:
                cv.imwrite(self.get_mask_path(self.curr_sample_idx), self.mask)
            else:
                for label_idx in range(min(9, nb_labels)):
                    if cls.CURRENT_KEY == pynput.keyboard.KeyCode(char=str(label_idx + 1)):
                        self.curr_label_idx = label_idx
                        curr_label = self.labels[self.curr_label_idx]
                        logger.debug("swapping to label #%s : '%s'" % (curr_label["id"], curr_label["name"]))
                        break
            cls.CURRENT_KEY = -1
        return False

    def get_mask_path(self, index):
        """Returns the path where the mask of a specific sample should be located."""
        dataset_name = None
        for name, offset in reversed(list(self.sample_idx_offsets.items())):
            if 0 <= index - offset < len(self.datasets[name]):
                index -= offset
                dataset_name = name
                break
        if dataset_name is None:
            raise AssertionError("bad logic somewhere")
        return os.path.join(self.annotations_dir, dataset_name, "%06d.png" % index)

    def load(self, index):
        """Loads the image and mask associated to a specific sample."""
        self.image = None
        for name, offset in reversed(list(self.sample_idx_offsets.items())):
            if 0 <= index - offset < len(self.datasets[name]):
                self.sample = self.datasets[name][index - offset]
                if self.sample_input_key not in self.sample:
                    raise AssertionError("could not locate value #%d for input key '%s'" % (index, self.sample_input_key))
                self.image = self.sample[self.sample_input_key]
                break
        if self.image is None or not isinstance(self.image, np.ndarray):
            raise AssertionError("invalid input image for index #%d" % index)
        self.image_display_base, self.image_display = None, None
        mask_path = self.get_mask_path(index)
        if os.path.isfile(mask_path):
            self.mask = cv.imread(mask_path, flags=cv.IMREAD_GRAYSCALE)
            if self.mask is None:
                raise AssertionError("could not open mask at '%s'" % mask_path)
        else:
            self.mask = np.full((self.image.shape[0], self.image.shape[1]), self.default_fill_id, dtype=np.uint8)
        self.mask_display = None
        if self.image.shape[0:2] != self.mask.shape:
            raise AssertionError("mismatched image/mask shapes")
        cls = ImageSegmentAnnotator
        if isinstance(self.window_size, str):
            cls.WINDOW_SIZE = thelper.utils.str2size(self.window_size)
        elif isinstance(self.window_size, int):
            max_scale_factor = self.window_size / max(self.image.shape[0], self.image.shape[1])
            cls.WINDOW_SIZE = (int(round(max_scale_factor * self.image.shape[1])), int(round(max_scale_factor * self.image.shape[0])))
        else:
            raise AssertionError("unexpected window size type")
        return self.image, self.mask

    def run(self):
        """Displays the main window and other GUI elements in a loop until it is closed by the user."""
        while cv.getWindowProperty(self.window_name, 0) != -1:
            if self.handle_keys():
                break
            self.brush.refresh(self.mask, self.labels[self.curr_label_idx])
            self.refresh_layers()
            self.refresh_gui()
        cv.destroyAllWindows()



"""Classification trainer/evaluator implementation module."""
import logging

import torch
import torch.optim

import thelper.utils
from thelper.train.base import Trainer

logger = logging.getLogger(__name__)


class ImageClassifTrainer(Trainer):
    """Trainer interface specialized for image classification.

    This class implements the abstract functions of :class:`thelper.train.base.Trainer` required to train/evaluate
    a model for image classification or recognition. It also provides a utility function for fetching i/o packets
    (images, class labels) from a sample, and that converts those into tensors for forwarding and loss estimation.

    .. seealso::
        | :class:`thelper.train.base.Trainer`
    """

    def __init__(self, session_name, save_dir, model, loaders, config, ckptdata=None):
        """Receives session parameters, parses image/label keys from task object, and sets up metrics."""
        super().__init__(session_name, save_dir, model, loaders, config, ckptdata=ckptdata)
        if not isinstance(self.model.task, thelper.tasks.Classification):
            raise AssertionError("expected task to be classification")
        self.input_key = self.model.task.get_input_key()
        self.label_key = self.model.task.get_gt_key()
        self.class_names = self.model.task.get_class_names()
        self.meta_keys = self.model.task.get_meta_keys()
        self.class_idxs_map = self.model.task.get_class_idxs_map()
        metrics = list(self.train_metrics.values()) + list(self.valid_metrics.values()) + list(self.test_metrics.values())
        for metric in metrics:  # check all metrics for classification-specific attributes, and set them
            if hasattr(metric, "set_class_names") and callable(metric.set_class_names):
                metric.set_class_names(self.class_names)
        self.warned_no_shuffling_augments = False

    def _to_tensor(self, sample):
        """Fetches and returns tensors of input images and class labels from a batched sample dictionary."""
        if not isinstance(sample, dict):
            raise AssertionError("trainer expects samples to come in dicts for key-based usage")
        if self.input_key not in sample:
            raise AssertionError("could not find input key '%s' in sample dict" % self.input_key)
        input_val, label_idx = sample[self.input_key], None
        if isinstance(input_val, list):
            if self.label_key in sample and sample[self.label_key] is not None:
                label = sample[self.label_key]
                if not isinstance(label, list) or len(label) != len(input_val):
                    raise AssertionError("label should also be a list of the same length as input")
                label_idx = [None] * len(input_val)
                for idx in range(len(input_val)):
                    input_val[idx], label_idx[idx] = self._to_tensor({self.input_key: input_val[idx],
                                                                      self.label_key: label[idx]})
            else:
                for idx in range(len(input_val)):
                    input_val[idx] = torch.FloatTensor(input_val[idx])
        else:
            input_val = torch.FloatTensor(input_val)
            if self.label_key in sample and sample[self.label_key] is not None:
                label = sample[self.label_key]
                if isinstance(label, torch.Tensor) and label.numel() == input_val.shape[0] \
                        and label.dtype == torch.int64:
                    label_idx = label  # shortcut with less checks (dataset is already using tensor'd indices)
                else:
                    label_idx = label_idx or list()
                    for class_name in label:
                        if isinstance(class_name, (int, torch.Tensor)):
                            if isinstance(class_name, torch.Tensor):
                                if torch.numel(class_name) != 1:
                                    raise AssertionError("unexpected scalar label, got vector")
                                class_name = class_name.item()
                            # dataset must already be using indices, we will forgive this...
                            if class_name < 0 or class_name >= len(self.class_names):
                                raise AssertionError("class name given as out-of-range index (%d) "
                                                     "for class list" % class_name)
                            class_name = self.class_names[class_name]
                        elif not isinstance(class_name, str):
                            raise AssertionError("expected label to be in str format "
                                                 "(task will convert to proper index)")
                        if class_name not in self.class_names:
                            raise AssertionError("got unexpected label '%s' for a sample (unknown class)" % class_name)
                        label_idx.append(self.class_idxs_map[class_name])
                    label_idx = torch.LongTensor(label_idx)
        return input_val, label_idx

    def _train_epoch(self, model, epoch, iter, dev, loss, optimizer, loader, metrics, monitor=None, writer=None):
        """Trains the model for a single epoch using the provided objects.

        Args:
            model: the model to train that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            iter: the iteration count at the start of the current epoch.
            dev: the target device that tensors should be uploaded to.
            loss: the loss function used to evaluate model fidelity.
            optimizer: the optimizer used for back propagation.
            loader: the data loader used to get transformed training samples.
            metrics: the list of metrics to evaluate after every iteration.
            monitor: name of the metric to update/monitor for improvements.
            writer: the writer used to store tbx events/messages/metrics.
        """
        if not loss:
            raise AssertionError("missing loss function")
        if not optimizer:
            raise AssertionError("missing optimizer")
        if not loader:
            raise AssertionError("no available data to load")
        if not isinstance(metrics, dict):
            raise AssertionError("expect metrics as dict object")
        epoch_loss = 0
        epoch_size = len(loader)
        self.logger.debug("fetching data loader samples...")
        for idx, sample in enumerate(loader):
            input_val, label = self._to_tensor(sample)
            optimizer.zero_grad()
            if label is None:
                raise AssertionError("groundtruth required when training a model")
            if isinstance(input_val, list):  # training samples got augmented, we need to backprop in multiple steps
                if not input_val:
                    raise AssertionError("cannot train with empty post-augment sample lists")
                if not isinstance(label, list) or len(label) != len(input_val):
                    raise AssertionError("label should also be a list of the same length as input")
                if not self.warned_no_shuffling_augments:
                    self.logger.warning("using training augmentation without global shuffling, "
                                        "gradient steps might be affected")
                    # see the docstring of thelper.transforms.operations.Duplicator for more information
                    self.warned_no_shuffling_augments = True
                iter_loss = None
                iter_pred = None
                augs_count = len(input_val)
                for input_idx in range(augs_count):
                    aug_pred = model(self._upload_tensor(input_val[input_idx], dev))
                    aug_loss = loss(aug_pred, self._upload_tensor(label[input_idx], dev))
                    aug_loss.backward()  # test backprop all at once? might not fit in memory...
                    if iter_pred is None:
                        iter_loss = aug_loss.clone().detach()
                        iter_pred = aug_pred.clone().detach()
                    else:
                        iter_loss += aug_loss.detach()
                        iter_pred = torch.cat((aug_pred.detach(), iter_pred), dim=0)
                iter_loss /= augs_count
                label = torch.cat(label, dim=0)
            else:
                iter_pred = model(self._upload_tensor(input_val, dev))
                iter_loss = loss(iter_pred, self._upload_tensor(label, dev))
                iter_loss.backward()
            if metrics:
                meta = {key: sample[key] if key in sample else None
                        for key in self.meta_keys} if self.meta_keys else None
                for metric in metrics.values():
                    metric.accumulate(iter_pred.detach().cpu(), label.detach().cpu(), meta=meta)
            if self.train_iter_callback is not None:
                self.train_iter_callback(sample=sample, task=self.model.task, pred=iter_pred,
                                         iter_idx=iter, max_iters=epoch_size,
                                         epoch_idx=epoch, max_epochs=self.epochs)
            epoch_loss += iter_loss.item()
            optimizer.step()
            monitor_output = ""
            if monitor is not None and monitor in metrics:
                monitor_output = "   {}: {:.2f}".format(monitor, metrics[monitor].eval())
            self.logger.info(
                "train epoch#{}  (iter#{})   batch: {}/{} ({:.0f}%)   loss: {:.6f}{}".format(
                    epoch,
                    iter,
                    idx + 1,
                    epoch_size,
                    ((idx + 1) / epoch_size) * 100.0,
                    iter_loss.item(),
                    monitor_output
                )
            )
            if writer:
                writer.add_scalar("iter/loss", iter_loss.item(), iter)
                for metric_name, metric in metrics.items():
                    if metric.is_scalar():  # only useful assuming that scalar metrics are smoothed...
                        writer.add_scalar("iter/%s" % metric_name, metric.eval(), iter)
            iter += 1
        epoch_loss /= epoch_size
        return epoch_loss, iter

    def _eval_epoch(self, model, epoch, dev, loader, metrics, monitor=None, writer=None):
        """Evaluates the model using the provided objects.

        Args:
            model: the model to evaluate that is already uploaded to the target device(s).
            epoch: the epoch number we are evaluating for (0-based).
            dev: the target device that tensors should be uploaded to.
            loader: the data loader used to get transformed valid/test samples.
            metrics: the dictionary of metrics to update every iteration.
            monitor: name of the metric to update/monitor for improvements.
            writer: the writer used to store tbx events/messages/metrics.
        """
        if not loader:
            raise AssertionError("no available data to load")
        with torch.no_grad():
            epoch_size = len(loader)
            self.logger.debug("fetching data loader samples...")
            for idx, sample in enumerate(loader):
                if idx < self.skip_eval_iter:
                    continue  # skip until previous iter count (if set externally; no effect otherwise)
                input_val, label = self._to_tensor(sample)
                if isinstance(input_val, list):  # evaluation samples got augmented, we need to get the mean prediction
                    if not input_val:
                        raise AssertionError("cannot eval with empty post-augment sample lists")
                    if not isinstance(label, list) or len(label) != len(input_val):
                        raise AssertionError("label should also be a list of the same length as input")
                    # this might be costly for nothing, we could remove the check and assume user is not dumb
                    if any([not torch.eq(l, label[0]).all() for l in label]):
                        raise AssertionError("all labels should be identical! (why do eval-time augment otherwise?)")
                    label = label[0]  # since all identical, just pick the first one and pretend its the only one
                    preds = None
                    for input_idx in range(len(input_val)):
                        pred = model(self._upload_tensor(input_val[input_idx], dev))
                        if preds is None:
                            preds = torch.unsqueeze(pred.clone(), 0)
                        else:
                            preds = torch.cat((preds, torch.unsqueeze(pred, 0)), 0)
                    pred = torch.mean(preds, dim=0)
                else:
                    pred = model(self._upload_tensor(input_val, dev))
                if metrics:
                    meta = {key: sample[key] if key in sample else None
                            for key in self.meta_keys} if self.meta_keys else None
                    for metric in metrics.values():
                        metric.accumulate(pred.cpu(), label.cpu() if label is not None else None, meta=meta)
                if self.eval_iter_callback is not None:
                    self.eval_iter_callback(sample=sample, task=self.model.task, pred=pred,
                                            iter_idx=idx, max_iters=epoch_size,
                                            epoch_idx=epoch, max_epochs=self.epochs)
                self.logger.info(
                    "eval epoch#{}   batch: {}/{} ({:.0f}%){}".format(
                        epoch,
                        idx + 1,
                        epoch_size,
                        ((idx + 1) / epoch_size) * 100.0,
                        "   {}: {:.2f}".format(monitor, metrics[monitor].eval()) if monitor is not None else ""
                    )
                )



# noinspection PyPackageRequirements
import numpy as np

import thelper
import thelper.transforms


def test_transform_custom_step_composers():
    transforms2 = thelper.transforms.CustomStepCompose(milestones={
        0: thelper.transforms.Resize(dsize=(16, 16)),
        2: thelper.transforms.Resize(dsize=(32, 32)),
        4: thelper.transforms.Resize(dsize=(64, 64)),
        8: thelper.transforms.Resize(dsize=(112, 112)),
        12: thelper.transforms.Resize(dsize=(160, 160)),
        15: thelper.transforms.Resize(dsize=(196, 196)),
        18: thelper.transforms.Resize(dsize=(224, 224)),
    }), thelper.transforms.CustomStepCompose(milestones={
        2: thelper.transforms.Resize(dsize=(32, 32)),
        4: thelper.transforms.Resize(dsize=(64, 64)),
        8: thelper.transforms.Resize(dsize=(112, 112)),
        12: thelper.transforms.Resize(dsize=(160, 160)),
        15: thelper.transforms.Resize(dsize=(196, 196)),
        18: thelper.transforms.Resize(dsize=(224, 224)),
    })
    for transforms in transforms2:
        for epoch in range(100):
            transforms.set_epoch(epoch)
            img = np.zeros((np.random.randint(1, 1000), np.random.randint(1, 1000), 1), dtype=np.uint8)
            img_t = transforms(img)
            closest_k = None
            for k, v in transforms.milestones.items():
                if k <= epoch:
                    closest_k = k
                else:
                    break
            if closest_k is None:
                assert img_t.shape == img.shape
            else:
                assert img_t.shape[:2] == transforms.milestones[closest_k].dsize
            if epoch >= 18:
                assert img_t.shape == (224, 224, 1)



# inspired from https://github.com/meetshah1995/pytorch-semseg/blob/master/ptsemseg/models/fcn.py

import numpy as np
import torch
import torch.nn
import torch.nn.functional

import thelper.nn


def get_upsampling_weight(in_channels, out_channels, kernel_size):
    factor = (kernel_size + 1) // 2
    if kernel_size % 2 == 1:
        center = factor - 1
    else:
        center = factor - 0.5
    og = np.ogrid[:kernel_size, :kernel_size]
    filt = (1 - abs(og[0] - center) / factor) * (1 - abs(og[1] - center) / factor)
    weight = np.zeros((in_channels, out_channels, kernel_size, kernel_size), dtype=np.float32)
    weight[list(range(in_channels)), list(range(out_channels)), :, :] = filt
    return torch.from_numpy(weight).float()


class FCN32s(thelper.nn.Module):
    def __init__(self, task, init_vgg16=True):
        super().__init__(task)
        self.conv_block1 = torch.nn.Sequential(
            torch.nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=100),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, stride=2, ceil_mode=True),
        )  # res = 1/2
        self.conv_block2 = torch.nn.Sequential(
            torch.nn.Conv2d(64, 128, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(128, 128, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, stride=2, ceil_mode=True),
        )  # res = 1/4
        self.conv_block3 = torch.nn.Sequential(
            torch.nn.Conv2d(128, 256, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, stride=2, ceil_mode=True),
        )  # res = 1/8
        self.conv_block4 = torch.nn.Sequential(
            torch.nn.Conv2d(256, 512, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(512, 512, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(512, 512, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, stride=2, ceil_mode=True),
        )  # res = 1/16
        self.conv_block5 = torch.nn.Sequential(
            torch.nn.Conv2d(512, 512, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(512, 512, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(512, 512, kernel_size=3, stride=1, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, stride=2, ceil_mode=True),
        )  # res = 1/32
        self.classifier = None
        self.upscaler = None
        self.set_task(task)
        if init_vgg16:
            import torchvision
            vgg16 = torchvision.models.vgg16(pretrained=True)
            self.init_vgg16_params(vgg16)

    def forward(self, x):
        conv = self.conv_block1(x)
        conv = self.conv_block2(conv)
        conv = self.conv_block3(conv)
        conv = self.conv_block4(conv)
        conv = self.conv_block5(conv)
        score = self.classifier(conv)
        out = self.upscaler(score)
        out = out[:, :, 19:(19 + x.size()[2]), 19:(19 + x.size()[3])].contiguous()
        #out = torch.nn.functional.upsample(score, x.size()[2:], mode="bilinear")
        return out

    def set_task(self, task):
        if isinstance(task, thelper.tasks.Segmentation):
            num_classes = len(task.get_class_names())
            if self.classifier is None or list(self.classifier.modules())[-1].out_channels != num_classes:
                self.classifier = torch.nn.Sequential(
                    torch.nn.Conv2d(512, 4096, kernel_size=7, stride=1, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout2d(),
                    torch.nn.Conv2d(4096, 4096, kernel_size=1, stride=1, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout2d(),
                    torch.nn.Conv2d(4096, num_classes, kernel_size=1, stride=1, padding=0),
                )
                self.upscaler = torch.nn.ConvTranspose2d(num_classes, num_classes,
                                                         kernel_size=64, stride=32, bias=False)
                self.upscaler.weight.data.copy_(get_upsampling_weight(num_classes, num_classes, 64))
        else:
            raise AssertionError("missing impl for non-segm task type")
        self.task = task

    def init_vgg16_params(self, vgg16, copy_fc8=True):
        blocks = [
            self.conv_block1,
            self.conv_block2,
            self.conv_block3,
            self.conv_block4,
            self.conv_block5,
        ]
        ranges = [[0, 4], [5, 9], [10, 16], [17, 23], [24, 29]]
        features = list(vgg16.features.children())
        for idx, conv_block in enumerate(blocks):
            for l1, l2 in zip(features[ranges[idx][0]: ranges[idx][1]], conv_block):
                if isinstance(l1, torch.nn.Conv2d) and isinstance(l2, torch.nn.Conv2d):
                    assert l1.weight.size() == l2.weight.size()
                    assert l1.bias.size() == l2.bias.size()
                    l2.weight.data = l1.weight.data
                    l2.bias.data = l1.bias.data
        for i1, i2 in zip([0, 3], [0, 3]):
            l1 = vgg16.classifier[i1]
            l2 = self.classifier[i2]
            l2.weight.data = l1.weight.data.view(l2.weight.size())
            l2.bias.data = l1.bias.data.view(l2.bias.size())
        n_class = self.classifier[6].weight.size()[0]
        if copy_fc8:
            l1 = vgg16.classifier[6]
            l2 = self.classifier[6]
            l2.weight.data = l1.weight.data[:n_class, :].view(l2.weight.size())
            l2.bias.data = l1.bias.data[:n_class]



"""Regression task interface module.

This module contains a class that defines the objectives of models/trainers for regression tasks.
"""
import logging

import numpy as np

from thelper.tasks.utils import Task

logger = logging.getLogger(__name__)


class Regression(Task):
    """Interface for n-dimension regression tasks.

    This specialization requests that when given an input tensor, the trained model should
    provide an n-dimensional target prediction. This is a fairly generic task that (unlike
    image classification and semantic segmentation) is not linked to a pre-existing set of
    possible solutions. The task interface is used to carry useful metadata for this task,
    e.g. input/output shapes, types, and min/max values for rounding/saturation.

    Attributes:
        input_shape: a numpy-compatible shape to expect model inputs to be in.
        target_shape: a numpy-compatible shape to expect the predictions to be in.
        target_type: a numpy-compatible type to cast the predictions to (if needed).
        target_min: an n-dim tensor containing minimum target values (if applicable).
        target_max: an n-dim tensor containing maximum target values (if applicable).
        input_key: the key used to fetch input tensors from a sample dictionary.
        target_key: the key used to fetch target (groundtruth) values from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.train.regr.RegressionTrainer`
    """

    def __init__(self, input_key, target_key, meta_keys=None, input_shape=None,
                 target_shape=None, target_type=None, target_min=None, target_max=None):
        """Receives and stores the target specs, input tensor key, and the extra (meta) keys produced by the dataset parser(s)."""
        super().__init__(input_key, target_key, meta_keys)
        if target_min is not None:
            if isinstance(target_min, (list, tuple)):
                target_min = np.asarray(target_min)
            if not isinstance(target_min, np.ndarray):
                raise AssertionError("target_min should be passed as list/tuple/ndarray")
        if target_max is not None:
            if isinstance(target_max, (list, tuple)):
                target_max = np.asarray(target_max)
            if not isinstance(target_max, np.ndarray):
                raise AssertionError("target_max should be passed as list/tuple/ndarray")
        if target_type is not None:
            if isinstance(target_min, np.ndarray) and target_min.dtype != target_type:
                raise AssertionError("invalid target min dtype")
            if isinstance(target_max, np.ndarray) and target_max.dtype != target_type:
                raise AssertionError("invalid target max dtype")
        if target_shape is not None:
            if isinstance(target_min, np.ndarray) and target_min.shape != target_shape:
                raise AssertionError("invalid target min shape")
            if isinstance(target_max, np.ndarray) and target_max.shape != target_shape:
                raise AssertionError("invalid target max shape")
        if isinstance(target_min, np.ndarray) and isinstance(target_max, np.ndarray):
            if target_min.shape != target_max.shape:
                raise AssertionError("target min/max shape mismatch")
        self.input_shape = input_shape
        if self.input_shape is not None:
            if isinstance(self.input_shape, list):
                self.input_shape = tuple(self.input_shape)
            if not isinstance(self.input_shape, tuple) or not all([isinstance(v, int) for v in self.input_shape]):
                raise AssertionError("unexpected input shape type (should be tuple of integers)")
        self.target_shape = target_shape
        if self.target_shape is not None:
            if isinstance(self.target_shape, list):
                self.target_shape = tuple(self.target_shape)
            if not isinstance(self.target_shape, tuple) or not all([isinstance(v, int) for v in self.target_shape]):
                raise AssertionError("unexpected target shape type (should be tuple of integers)")
        self.target_type = target_type
        if self.target_type is not None:
            if isinstance(self.target_type, str):
                import thelper.utils
                self.target_type = thelper.utils.import_class(self.target_type)
            if not issubclass(self.target_type, np.generic):
                raise AssertionError("target type should be a numpy-compatible type")
        self.target_min = target_min
        self.target_max = target_max

    def get_input_shape(self):
        """Returns the shape of inputs to be processed by the model."""
        return self.input_shape

    def get_target_shape(self):
        """Returns the shape of outputs to be predicted by the model."""
        return self.target_shape

    def get_target_type(self):
        """Returns the type of outputs to be predicted by the model."""
        return self.target_type

    def get_target_min(self):
        """Returns the minimum target value(s) to be predicted by the model."""
        return self.target_min

    def get_target_max(self):
        """Returns the maximum target value(s) to be predicted by the model."""
        return self.target_max

    def check_compat(self, other, exact=False):
        """Returns whether the current task is compatible with the provided one or not.

        This is useful for sanity-checking, and to see if the inputs/outputs of two models
        are compatible. If ``exact = True``, all fields will be checked for exact (perfect)
        compatibility (in this case, matching meta keys).
        """
        if isinstance(other, Regression):
            # if both tasks are related to regression: all non-None keys and specs must match
            return (self.get_input_key() == other.get_input_key() and
                    (self.get_gt_key() is None or other.get_gt_key() is None or
                     self.get_gt_key() == other.get_gt_key()) and
                    (self.get_input_shape() is None or other.get_input_shape() is None or
                     self.get_input_shape() == other.get_input_shape()) and
                    (self.get_target_shape() is None or other.get_target_shape() is None or
                     self.get_target_shape() == other.get_target_shape()) and
                    (self.get_target_type() is None or other.get_target_type() is None or
                     self.get_target_type() == other.get_target_type()) and
                    (self.get_target_min() is None or other.get_target_min() is None or
                     self.get_target_min() == other.get_target_min()) and
                    (self.get_target_max() is None or other.get_target_max() is None or
                     self.get_target_max() == other.get_target_max()) and
                    (not exact or (self.get_meta_keys() == other.get_meta_keys())))
        elif type(other) == Task:
            # if 'other' simply has no gt, compatibility rests on input key only
            return not exact and self.get_input_key() == other.get_input_key() and other.get_gt_key() is None
        return False

    def get_compat(self, other):
        """Returns a task instance compatible with the current task and the given one."""
        if not self.check_compat(other):
            raise AssertionError("cannot create compatible task instance between:\n"
                                 "\tself: %s\n\tother: %s" % (str(self), str(other)))
        meta_keys = list(set(self.get_meta_keys() + other.get_meta_keys()))
        return Regression(self.get_input_key(), self.get_gt_key(),
                          meta_keys=meta_keys, input_shape=self.get_input_shape(),
                          target_shape=self.get_target_shape(), target_type=self.get_target_type(),
                          target_min=self.get_target_min(), target_max=self.get_target_max())

    def __repr__(self):
        """Creates a print-friendly representation of a segmentation task."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + ": " + str({
            "input_key": self.get_input_key(),
            "target_key": self.get_gt_key(),
            "meta_keys": self.get_meta_keys(),
            "input_shape": self.get_input_shape(),
            "target_shape": self.get_target_shape(),
            "target_type": self.get_target_type(),
            "target_min": self.get_target_min(),
            "target_max": self.get_target_max()
        })



"""Optimization/metrics package.

This package contains metrics implementations used to monitor training sessions and evaluate models,
and optimization methods used to control the learning behavior of these models.
"""

import logging

import thelper.optim.metrics  # noqa: F401
import thelper.optim.schedulers  # noqa: F401
import thelper.optim.utils  # noqa: F401
from thelper.optim.metrics import BinaryAccuracy  # noqa: F401
from thelper.optim.metrics import CategoryAccuracy  # noqa: F401
from thelper.optim.metrics import ClassifLogger  # noqa: F401
from thelper.optim.metrics import ClassifReport  # noqa: F401
from thelper.optim.metrics import ConfusionMatrix  # noqa: F401
from thelper.optim.metrics import ExternalMetric  # noqa: F401
from thelper.optim.metrics import MeanAbsoluteError  # noqa: F401
from thelper.optim.metrics import MeanSquaredError  # noqa: F401
from thelper.optim.metrics import Metric  # noqa: F401
from thelper.optim.metrics import RawPredictions  # noqa: F401
from thelper.optim.metrics import ROCCurve  # noqa: F401
from thelper.optim.schedulers import CustomStepLR  # noqa: F401
from thelper.optim.utils import create_loss_fn  # noqa: F401
from thelper.optim.utils import create_metrics  # noqa: F401
from thelper.optim.utils import create_optimizer  # noqa: F401
from thelper.optim.utils import create_scheduler  # noqa: F401
from thelper.optim.utils import get_lr  # noqa: F401

logger = logging.getLogger("thelper.optim")



"""Transformation operations package.

This package contains data transformation classes and wrappers for preprocessing,
augmentation, and normalization of data samples.
"""

import logging

import thelper.transforms.operations  # noqa: F401
import thelper.transforms.utils  # noqa: F401
import thelper.transforms.wrappers  # noqa: F401
from thelper.transforms.operations import Affine  # noqa: F401
from thelper.transforms.operations import CenterCrop  # noqa: F401
from thelper.transforms.operations import Compose  # noqa: F401
from thelper.transforms.operations import CustomStepCompose  # noqa: F401
from thelper.transforms.operations import Duplicator  # noqa: F401
from thelper.transforms.operations import NormalizeMinMax  # noqa: F401
from thelper.transforms.operations import NormalizeZeroMeanUnitVar  # noqa: F401
from thelper.transforms.operations import RandomResizedCrop  # noqa: F401
from thelper.transforms.operations import RandomShift  # noqa: F401
from thelper.transforms.operations import Resize  # noqa: F401
from thelper.transforms.operations import Tile  # noqa: F401
from thelper.transforms.operations import ToNumpy  # noqa: F401
from thelper.transforms.operations import Transpose  # noqa: F401
from thelper.transforms.utils import load_augments  # noqa: F401
from thelper.transforms.utils import load_transforms  # noqa: F401
from thelper.transforms.wrappers import AlbumentationsWrapper  # noqa: F401
from thelper.transforms.wrappers import AugmentorWrapper  # noqa: F401
from thelper.transforms.wrappers import TransformWrapper  # noqa: F401

logger = logging.getLogger("thelper.transforms")



"""Dataset parsers module.

This module contains dataset parser interfaces and base classes that define basic i/o
operations so that the framework can automatically interact with training data.
"""

import logging
import os
from abc import abstractmethod

import cv2 as cv
import numpy as np
import PIL
import PIL.Image
import torch
import torch.utils.data

import thelper.tasks
import thelper.utils

logger = logging.getLogger(__name__)


class Dataset(torch.utils.data.Dataset):
    """Abstract dataset parsing interface that holds a task and a list of sample dictionaries.

    This interface helps fix a failure of PyTorch's dataset interface (``torch.utils.data.Dataset``):
    the lack of identity associated with the components of a sample. In short, a data sample loaded by a
    dataset typically contains the input data that should be forwarded to a model as well as the expected
    prediction of the model (i.e. the 'groundtruth') that will be used to compute the loss. These two
    elements are typically paired in a tuple that can then be provided to the data loader for batching.
    Problems however arise when the model has multiple inputs or outputs, when the sample needs to carry
    supplemental metadata to simplify debugging, or when transformation operations need to be applied
    only to specific elements of the sample. Here, we fix this issue by specifying that all samples must
    be provided to data loaders as dictionaries. The keys of these dictionaries explicitly define which
    value(s) should be transformed, which should be forwarded to the model, which are the expected model
    predictions, and which are only used for debugging. The keys are defined via the task object that is
    generated by the dataset or specified via the configuration file (see :class:`thelper.tasks.utils.Task`
    for more information).

    To properly use this interface, a derived class must thus implement :func:`thelper.data.parsers.Dataset.__getitem__`,
    :func:`thelper.data.parsers.Dataset.get_task`, and store its samples as dictionaries in ``self.samples``.

    Attributes:
        config: dictionary of extra parameters that are required by the dataset interface.
        transforms: function or object that should be applied to all loaded samples in order to
            return the data in the requested transformed/augmented state.
        deepcopy: specifies whether this dataset interface should be deep-copied inside
            :func:`thelper.data.utils._LoaderFactory.create_loaders` so that it may be shared between
            different threads. This is false by default, as we assume datasets do not contain a state
            or buffer that might cause problems in multi-threaded data loaders.
        samples: list of dictionaries containing the data that is ready to be forwarded to the
            data loader. Note that relatively costly operations (such as reading images from a disk
            or pre-transforming them) should be delayed until the :func:`thelper.data.parsers.Dataset.__getitem__`
            function is called, as they will most likely then be accomplished in a separate thread.
            Once loaded, these samples should never be modified by another part of the framework. For
            example, transformation and augmentation operations will always be applied to copies
            of these samples.

    .. seealso::
        | :class:`thelper.data.parsers.ExternalDataset`
    """

    def __init__(self, config=None, transforms=None, deepcopy=False):
        """Dataset parser constructor.

        In order for derived datasets to be instantiated automatically be the framework from a
        configuration file, the signature of their constructors should match the one shown here.
        This means all required extra parameters must be passed in the 'config' argument, which is
        a dictionary.

        Args:
            config: dictionary of extra parameters that are required by the dataset interface.
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :func:`thelper.data.utils._LoaderFactory.create_loaders` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super().__init__()
        self.config = config
        self.transforms = transforms
        self.deepcopy = deepcopy  # will determine if we deepcopy in each loader
        self.samples = None  # must be filled by the derived class as a list of dictionaries

    def _get_derived_name(self):
        """Returns a pretty-print version of the derived class's name."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__

    def __len__(self):
        """Returns the total number of samples available from this dataset interface."""
        return len(self.samples)

    def __iter__(self):
        """Returns an iterator over the dataset's samples."""
        for idx in range(len(self.samples)):
            yield self[idx]

    @abstractmethod
    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        raise NotImplementedError

    @abstractmethod
    def get_task(self):
        """Returns the dataset task object that provides the i/o keys for parsing sample dicts."""
        raise NotImplementedError

    def __repr__(self):
        """Returns a print-friendly representation of this dataset."""
        return self._get_derived_name() + ": {{\n\tsize: {},\n\tdeepcopy: {},\n\ttransforms: {}\n}}".format(
            str(len(self)), str(self.deepcopy), str(self.transforms)
        )


class ClassificationDataset(Dataset):
    """Classification dataset specialization interface.

    This specialization receives some extra parameters in its constructor and automatically defines
    its task (:class:`thelper.tasks.classif.Classification`) based on those. The derived class must still
    implement :func:`thelper.data.parsers.ClassificationDataset.__getitem__`, and it must still store its
    samples as dictionaries in ``self.samples`` to behave properly.

    Attributes:
        task: classification task object containing the key information passed in the constructor.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, class_names, input_key, label_key, meta_keys=None, config=None,
                 transforms=None, deepcopy=False):
        """Classification dataset parser constructor.

        In order for derived datasets to be instantiated automatically by the framework from a
        configuration file, the signature of their constructors should match the one shown here.
        This means all required extra parameters must be passed in the 'config' argument, which is
        a dictionary.

        Args:
            class_names: list of all class names (or labels) that will be associated with the samples.
            input_key: key used to index the input data in the loaded samples.
            label_key: key used to index the label (or class name) in the loaded samples.
            meta_keys: list of extra keys that will be available in the loaded samples.
            config: dictionary of extra parameters that are required by the dataset interface.
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :func:`thelper.data.utils._LoaderFactory.create_loaders` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super().__init__(config=config, transforms=transforms, deepcopy=deepcopy)
        self.task = thelper.tasks.Classification(class_names, input_key, label_key, meta_keys=meta_keys)

    @abstractmethod
    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        raise NotImplementedError

    def get_task(self):
        """Returns the dataset task object that provides the i/o keys for parsing sample dicts."""
        return self.task


class SegmentationDataset(Dataset):
    """Segmentation dataset specialization interface.

    This specialization receives some extra parameters in its constructor and automatically defines
    its task (:class:`thelper.tasks.segm.Segmentation`) based on those. The derived class must still
    implement :func:`thelper.data.parsers.SegmentationDataset.__getitem__`, and it must still store its
    samples as dictionaries in ``self.samples`` to behave properly.

    Attributes:
        task: segmentation task object containing the key information passed in the constructor.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, class_names, input_key, label_map_key, meta_keys=None, dontcare=None,
                 config=None, transforms=None, deepcopy=False):
        """Segmentation dataset parser constructor.

        In order for derived datasets to be instantiated automatically by the framework from a
        configuration file, the signature of their constructors should match the one shown here.
        This means all required extra parameters must be passed in the 'config' argument, which is
        a dictionary.

        Args:
            class_names: list of all class names (or labels) that must be predicted in the image.
            input_key: key used to index the input image in the loaded samples.
            label_map_key: key used to index the label map in the loaded samples.
            meta_keys: list of extra keys that will be available in the loaded samples.
            config: dictionary of extra parameters that are required by the dataset interface.
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :func:`thelper.data.utils._LoaderFactory.create_loaders` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super().__init__(config=config, transforms=transforms, deepcopy=deepcopy)
        self.task = thelper.tasks.Segmentation(class_names, input_key, label_map_key,
                                               meta_keys=meta_keys, dontcare=dontcare)

    @abstractmethod
    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        raise NotImplementedError

    def get_task(self):
        """Returns the dataset task object that provides the i/o keys for parsing sample dicts."""
        return self.task


class ImageDataset(Dataset):
    """Image dataset specialization interface.

    This specialization is used to parse simple image folders, and it does not fulfill the requirements of any
    specialized task constructors due to the lack of groundtruth data support. Therefore, it returns a basic task
    object (:class:`thelper.tasks.utils.Task`) with no set value for the groundtruth key, and it cannot be used to
    directly train a model. It can however be useful when simply visualizing, annotating, or testing raw data
    from a simple directory structure.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, config=None, transforms=None):
        """Image dataset parser constructor.

        This baseline constructor matches the signature of :class:`thelper.data.parsers.Dataset`, and simply
        forwards its parameters.
        """
        super().__init__(config=config, transforms=transforms)
        self.root = thelper.utils.get_key("root", config)
        if self.root is None or not os.path.isdir(self.root):
            raise AssertionError("invalid input data root '%s'" % self.root)
        self.image_key = thelper.utils.get_key_def("image_key", config, "image")
        self.path_key = thelper.utils.get_key_def("path_key", config, "path")
        self.idx_key = thelper.utils.get_key_def("idx_key", config, "idx")
        self.samples = []
        for folder, subfolder, files in os.walk(self.root):
            for file in files:
                ext = os.path.splitext(file)[1].lower()
                if ext in [".jpg", ".jpeg", ".bmp", ".png", ".ppm", ".pgm", ".tif"]:
                    self.samples.append({self.path_key: os.path.join(folder, file)})
        self.task = thelper.tasks.Task(self.image_key, None, [self.path_key, self.idx_key])

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if idx < 0 or idx >= len(self.samples):
            raise AssertionError("sample index is out-of-range")
        sample = self.samples[idx]
        image_path = sample[self.path_key]
        image = cv.imread(image_path)
        if image is None:
            raise AssertionError("invalid image at '%s'" % image_path)
        sample = {
            self.image_key: image,
            self.idx_key: idx,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample

    def get_task(self):
        """Returns the dataset task object that provides the i/o keys for parsing sample dicts."""
        return self.task


class ImageFolderDataset(ClassificationDataset):
    """Image folder dataset specialization interface for classification tasks.

    This specialization is used to parse simple image subfolders, and it essentially replaces the very
    basic ``torchvision.datasets.ImageFolder`` interface with similar functionalities. It it used to provide
    a proper task interface as well as path metadata in each loaded packet for metrics/logging output.

    .. seealso::
        | :class:`thelper.data.parsers.ImageDataset`
        | :class:`thelper.data.parsers.ClassificationDataset`
    """

    def __init__(self, config=None, transforms=None):
        """Image folder dataset parser constructor."""
        self.root = thelper.utils.get_key("root", config)
        if self.root is None or not os.path.isdir(self.root):
            raise AssertionError("invalid input data root '%s'" % self.root)
        class_map = {}
        for child in os.listdir(self.root):
            if os.path.isdir(os.path.join(self.root, child)):
                class_map[child] = []
        if not class_map:
            raise AssertionError("could not find any image folders at '%s'" % self.root)
        image_exts = [".jpg", ".jpeg", ".bmp", ".png", ".ppm", ".pgm", ".tif"]
        self.image_key = thelper.utils.get_key_def("image_key", config, "image")
        self.path_key = thelper.utils.get_key_def("path_key", config, "path")
        self.idx_key = thelper.utils.get_key_def("idx_key", config, "idx")
        self.label_key = thelper.utils.get_key_def("label_key", config, "label")
        samples = []
        for class_name in class_map:
            class_folder = os.path.join(self.root, class_name)
            for folder, subfolder, files in os.walk(class_folder):
                for file in files:
                    ext = os.path.splitext(file)[1].lower()
                    if ext in image_exts:
                        class_map[class_name].append(len(samples))
                        samples.append({
                            self.path_key: os.path.join(folder, file),
                            self.label_key: class_name
                        })
        class_map = {k: v for k, v in class_map.items() if len(v) > 0}
        if not class_map:
            raise AssertionError("could not locate any subdir in '%s' with images to load" % self.root)
        meta_keys = [self.path_key, self.idx_key]
        super().__init__(class_names=list(class_map.keys()), input_key=self.image_key,
                         label_key=self.label_key, meta_keys=meta_keys, config=config, transforms=transforms)
        self.samples = samples

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if idx < 0 or idx >= len(self.samples):
            raise AssertionError("sample index is out-of-range")
        sample = self.samples[idx]
        image_path = sample[self.path_key]
        image = cv.imread(image_path)
        if image is None:
            raise AssertionError("invalid image at '%s'" % image_path)
        sample = {
            self.image_key: image,
            self.idx_key: idx,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample


class ExternalDataset(Dataset):
    """External dataset interface.

    This interface allows external classes to be instantiated automatically in the framework through
    a configuration file, as long as they themselves provide implementations for  ``__getitem__`` and
    ``__len__``. This includes all derived classes of ``torch.utils.data.Dataset`` such as
    ``torchvision.datasets.ImageFolder``, and the specialized versions such as ``torchvision.datasets.CIFAR10``.

    Note that for this interface to be compatible with our runtime instantiation rules, the constructor
    needs to receive a fully constructed task object. This object is currently constructed in
    :func:`thelper.data.parsers.create_parsers` based on extra parameters; see the code there for more
    information.

    Attributes:
        dataset_type: type of the external dataset object to instantiate
        task: task object containing the key information passed in the external configuration.
        samples: instantiation of the dataset object itself, faking the presence of a list of samples
        warned_dictionary: specifies whether the user was warned about missing keys in the output
            samples dictionaries.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    def __init__(self, dataset_type, task, config=None, transforms=None, deepcopy=False):
        """External dataset parser constructor.

        Args:
            dataset_type: fully qualified name of the dataset object to instantiate
            task: fully constructed task object providing key information for sample loading.
            config: dictionary of extra parameters that are required by the dataset interface.
            transforms: function or object that should be applied to all loaded samples in order to
                return the data in the requested transformed/augmented state.
            deepcopy: specifies whether this dataset interface should be deep-copied inside
                :func:`thelper.data.utils._LoaderFactory.create_loaders` so that it may be shared between
                different threads. This is false by default, as we assume datasets do not contain a state
                or buffer that might cause problems in multi-threaded data loaders.
        """
        super().__init__(config=config, transforms=transforms, deepcopy=deepcopy)
        if not dataset_type or not hasattr(dataset_type, "__getitem__") or not hasattr(dataset_type, "__len__"):
            raise AssertionError("external dataset type must implement '__getitem__' and '__len__' methods")
        if task is None or not isinstance(task, thelper.tasks.Task):
            raise AssertionError("task type must derive from thelper.tasks.Task")
        self.dataset_type = dataset_type
        self.task = task
        self.samples = dataset_type(**config)
        self.warned_dictionary = False

    def _get_derived_name(self):
        """Returns a pretty-print version of the external class's name."""
        return self.dataset_type.__module__ + "." + self.dataset_type.__qualname__

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        sample = self.samples[idx]
        if sample is None:
            # since might have provided an invalid sample count before, it's dangerous to skip empty samples here
            raise AssertionError("invalid sample received in external dataset impl")
        warn_dictionary = False
        if isinstance(sample, (list, tuple)):
            out_sample_list = []
            for idx, subs in enumerate(sample):
                if isinstance(subs, PIL.Image.Image):
                    subs = np.array(subs)
                out_sample_list.append(subs)
            sample = {str(idx): out_sample_list[idx] for idx in range(len(out_sample_list))}
            warn_dictionary = True
        elif isinstance(sample, (np.ndarray, PIL.Image.Image, torch.Tensor)):
            sample = {"0": sample}
            warn_dictionary = True
        if not isinstance(sample, dict):
            # could add checks to see if the sample already behaves like a dict? todo
            raise AssertionError("no clue how to convert given data sample into dictionary")
        if warn_dictionary and not self.warned_dictionary:
            logger.warning("dataset '%s' not returning samples as dictionaries;"
                           " will blindly map elements to their indices" % self._get_derived_name())
            self.warned_dictionary = True
        if self.transforms:
            sample = self.transforms(sample)
        return sample

    def get_task(self):
        """Returns the dataset task object that provides the i/o keys for parsing sample dicts."""
        return self.task



"""General utilities module.

This module only contains non-ML specific functions, i/o helpers,
and matplotlib/pyplot drawing calls.
"""
import copy
import errno
import functools
import glob
import importlib
import inspect
import io
import itertools
import json
import logging
import math
import os
import platform
import re
import sys
import time
from typing import AnyStr, Callable, List, Optional  # noqa: F401

import cv2 as cv
import matplotlib.pyplot as plt
import numpy as np
import PIL.Image
import sklearn.metrics
import torch

import thelper.typedefs  # noqa: F401

logger = logging.getLogger(__name__)
bypass_queries = False


class Struct(object):
    """Generic runtime-defined C-like data structure (maps constructor elements to fields)."""

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __repr__(self):
        return self.__class__.__name__ + ": " + str(self.__dict__)


def get_available_cuda_devices(attempts_per_device=5):
    # type: (Optional[int]) -> List[int]
    """
    Tests all visible cuda devices and returns a list of available ones.

    Returns:
        List of available cuda device IDs (integers). An empty list means no
        cuda device is available, and the app should fallback to cpu.
    """
    if not torch.cuda.is_available() or torch.cuda.device_count() == 0:
        return []
    devices_available = [False] * torch.cuda.device_count()
    attempt_broadcast = False
    for attempt in range(attempts_per_device):
        for device_id in range(torch.cuda.device_count()):
            if not devices_available[device_id]:
                if not attempt_broadcast:
                    logger.debug("testing availability of cuda device #%d (%s)" % (
                        device_id, torch.cuda.get_device_name(device_id)
                    ))
                # noinspection PyBroadException
                try:
                    torch.cuda.set_device(device_id)
                    test_val = torch.cuda.FloatTensor([1])
                    if test_val.cpu().item() != 1.0:
                        raise AssertionError("sometime's really wrong")
                    devices_available[device_id] = True
                except Exception:
                    pass
        attempt_broadcast = True
    return [device_id for device_id, available in enumerate(devices_available) if available]


# noinspection PyUnusedLocal
def setup_cv2(config):
    """Parses the provided config for OpenCV flags and sets up its global state accordingly."""
    # https://github.com/pytorch/pytorch/issues/1355
    cv.setNumThreads(0)
    cv.ocl.setUseOpenCL(False)
    # todo: add more global opencv flags setups here


def setup_cudnn(config):
    """Parses the provided config for CUDNN flags and sets up PyTorch accordingly."""
    if "cudnn" in config and isinstance(config["cudnn"], dict):
        config = config["cudnn"]
        if "benchmark" in config:
            cudnn_benchmark_flag = str2bool(config["benchmark"])
            logger.debug("cudnn benchmark mode = %s" % str(cudnn_benchmark_flag))
            torch.backends.cudnn.benchmark = cudnn_benchmark_flag
        if "deterministic" in config:
            cudnn_deterministic_flag = str2bool(config["deterministic"])
            logger.debug("cudnn deterministic mode = %s" % str(cudnn_deterministic_flag))
            torch.backends.cudnn.deterministic = cudnn_deterministic_flag
    else:
        if "cudnn_benchmark" in config:
            cudnn_benchmark_flag = str2bool(config["cudnn_benchmark"])
            logger.debug("cudnn benchmark mode = %s" % str(cudnn_benchmark_flag))
            torch.backends.cudnn.benchmark = cudnn_benchmark_flag
        if "cudnn_deterministic" in config:
            cudnn_deterministic_flag = str2bool(config["cudnn_deterministic"])
            logger.debug("cudnn deterministic mode = %s" % str(cudnn_deterministic_flag))
            torch.backends.cudnn.deterministic = cudnn_deterministic_flag


def setup_globals(config):
    """Parses the provided config for global flags and sets up the global state accordingly."""
    if "bypass_queries" in config and config["bypass_queries"]:
        global bypass_queries
        bypass_queries = True
    setup_cv2(config)
    setup_cudnn(config)


def load_checkpoint(ckpt,                      # type: thelper.typedefs.CheckpointLoadingType
                    map_location=None,         # type: Optional[thelper.typedefs.MapLocationType]
                    always_load_latest=False,  # type: Optional[bool]
                    check_version=True,        # type: Optional[bool]
                    ):                         # type: (...) -> thelper.typedefs.CheckpointContentType
    """Loads a session checkpoint via PyTorch, check its compatibility, and returns its data.

    If the ``ckpt`` parameter is a path to a valid directory, then that directly will be searched for
    a checkpoint. If multiple checkpoints are found, the latest will be returned (based on the epoch
    index in its name). iF ``always_load_latest`` is set to False and if a checkpoint named
    ``ckpt.best.pth`` is found, it will be returned instead.

    Args:
        ckpt: a file-like object or a path to the checkpoint file or session directory.
        map_location: a function, string or a dict specifying how to remap storage
            locations. See ``torch.load`` for more information.
        always_load_latest: toggles whether to always try to load the latest checkpoint
            if a session directory is provided (instead of loading the 'best' checkpoint).
        check_version: toggles whether the checkpoint's version should be checked for
            compatibility issues, and query the user for how to proceed.

    Returns:
        Content of the checkpoint (a dictionary).
    """
    if map_location is None and not get_available_cuda_devices():
        map_location = 'cpu'
    if isinstance(ckpt, str) and os.path.isdir(ckpt):
        logger.debug("will search directory '%s' for a checkpoint to load..." % ckpt)
        search_ckpt_dir = os.path.join(ckpt, "checkpoints")
        if os.path.isdir(search_ckpt_dir):
            search_dir = search_ckpt_dir
        else:
            search_dir = ckpt
        ckpt_paths = glob.glob(os.path.join(search_dir, "ckpt.*.pth"))
        if not ckpt_paths:
            raise AssertionError("could not find any valid checkpoint files in directory '%s'" % search_dir)
        latest_checkpoint_epoch = -1
        for ckpt_path in ckpt_paths:
            # note: the 2nd field in the name should be the epoch index, or 'best' if final checkpoint
            tag = os.path.basename(ckpt_path).split(".")[1]
            if tag == "best" and (not always_load_latest or latest_checkpoint_epoch == -1):
                # if eval-only, always pick the best checkpoint; otherwise, only pick if nothing else exists
                ckpt = ckpt_path
                if not always_load_latest:
                    break
            elif tag != "best" and int(tag) > latest_checkpoint_epoch:  # otherwise, pick latest
                # note: if several sessions are merged, this will pick the latest checkpoint of the first...
                ckpt = ckpt_path
                latest_checkpoint_epoch = int(tag)
        if not os.path.isfile(ckpt):
            raise AssertionError("could not find valid checkpoint at '%s'" % ckpt)
    if isinstance(ckpt, str):
        logger.debug("parsing checkpoint at '%s'" % ckpt)
    else:
        logger.debug("parsing checkpoint provided via file object")
    ckptdata = torch.load(ckpt, map_location=map_location)
    if not isinstance(ckptdata, dict):
        raise AssertionError("unexpected checkpoint data type")
    if check_version:
        good_version = False
        from thelper import __version__ as curr_ver
        if "version" not in ckptdata:
            logger.warning("checkpoint missing internal version tag")
            ckpt_ver_str = "0.0.0"
        else:
            ckpt_ver_str = ckptdata["version"]
            if not isinstance(ckpt_ver_str, str) or len(ckpt_ver_str.split(".")) != 3:
                raise AssertionError("unexpected checkpoint version formatting")
            # by default, checkpoints should be from the same minor version, we warn otherwise
            versions = [curr_ver.split("."), ckpt_ver_str.split(".")]
            if versions[0][0] != versions[1][0]:
                logger.error("incompatible checkpoint, major version mismatch (%s vs %s)" % (curr_ver, ckpt_ver_str))
            elif versions[0][1] != versions[1][1]:
                logger.warning("outdated checkpoint, minor version mismatch (%s vs %s)" % (curr_ver, ckpt_ver_str))
            else:
                good_version = True
        if not good_version:
            answer = query_string("Checkpoint version unsupported (framework=%s, checkpoint=%s); how do you want to proceed?" %
                                  (curr_ver, ckpt_ver_str), choices=["continue", "migrate", "abort"], default="migrate", bypass="migrate")
            if answer == "abort":
                logger.error("checkpoint out-of-date; user aborted")
                sys.exit(1)
            elif answer == "continue":
                logger.warning("will attempt to load checkpoint anyway (might crash later due to incompatibilities)")
            elif answer == "migrate":
                ckptdata = migrate_checkpoint(ckptdata)
    return ckptdata


def migrate_checkpoint(ckptdata,  # type: thelper.typedefs.CheckpointContentType
                       ):         # type: (...) -> thelper.typedefs.CheckpointContentType
    """Migrates the content of an incompatible or outdated checkpoint to the current version of the framework.

    This function might not be able to fix all backward compatibility issues (e.g. it cannot fix class interfaces
    that were changed). Perfect reproductibility of tests cannot be guaranteed either if this migration tool is used.

    Args:
        ckptdata: checkpoint data in dictionary form obtained via ``thelper.utils.load_checkpoint``. Note that
            the data contained in this dictionary will be modified in-place.

    Returns:
        An updated checkpoint dictionary that should be compatible with the current version of the framework.
    """
    if not isinstance(ckptdata, dict):
        raise AssertionError("unexpected ckptdata type")
    from thelper import __version__ as curr_ver
    curr_ver = [int(num) for num in curr_ver.split(".")]
    ckpt_ver_str = ckptdata["version"] if "version" in ckptdata else "0.0.0"
    ckpt_ver = [int(num) for num in ckpt_ver_str.split(".")]
    if (ckpt_ver[0] > curr_ver[0] or (ckpt_ver[0] == curr_ver[0] and ckpt_ver[1] > curr_ver[1]) or
       (ckpt_ver[0:2] == curr_ver[0:2] and ckpt_ver[2] > curr_ver[2])):
        raise AssertionError("cannot migrate checkpoints from future versions!")
    if "config" not in ckptdata:
        raise AssertionError("checkpoint migration requires config")
    old_config = ckptdata["config"]
    new_config = migrate_config(copy.deepcopy(old_config), ckpt_ver_str)
    if ckpt_ver == [0, 0, 0]:
        logger.warning("trying to migrate checkpoint data from v0.0.0; all bets are off")
    else:
        logger.info("trying to migrate checkpoint data from v%s" % ckpt_ver_str)
    if ckpt_ver[0] <= 0 and ckpt_ver[1] <= 1:
        # combine 'host' and 'time' fields into 'source'
        if "host" in ckptdata and "time" in ckptdata:
            ckptdata["source"] = ckptdata["host"] + ckptdata["time"]
            del ckptdata["host"]
            del ckptdata["time"]
        # update classif task interface
        if "task" in ckptdata and isinstance(ckptdata["task"], thelper.tasks.classif.Classification):
            ckptdata["task"] = str(thelper.tasks.classif.Classification(class_names=ckptdata["task"].class_names,
                                                                        input_key=ckptdata["task"].input_key,
                                                                        label_key=ckptdata["task"].label_key,
                                                                        meta_keys=ckptdata["task"].meta_keys))
        # move 'state_dict' field to 'model'
        if "state_dict" in ckptdata:
            ckptdata["model"] = ckptdata["state_dict"]
            del ckptdata["state_dict"]
        # create 'model_type' and 'model_params' fields
        if "model" in new_config:
            if "type" in new_config["model"]:
                ckptdata["model_type"] = new_config["model"]["type"]
            else:
                ckptdata["model_type"] = None
            if "params" in new_config["model"]:
                ckptdata["model_params"] = copy.deepcopy(new_config["model"]["params"])
            else:
                ckptdata["model_params"] = {}
        # TODO: create 'scheduler' field to restore previous state? (not so important for early versions)
        # ckpt_ver = [0, 2, 0]  # set ver for next update step
    # if ckpt_ver[0] <= x and ckpt_ver[1] <= y and ckpt_ver[2] <= z:
    #     ... add more compatibility fixes here
    ckptdata["config"] = new_config
    return ckptdata


def migrate_config(config,        # type: thelper.typedefs.ConfigDict
                   cfg_ver_str,   # type: str
                   ):             # type: (...) -> thelper.typedefs.ConfigDict
    """Migrates the content of an incompatible or outdated configuration to the current version of the framework.

    This function might not be able to fix all backward compatibility issues (e.g. it cannot fix class interfaces
    that were changed). Perfect reproductibility of tests cannot be guaranteed either if this migration tool is used.

    Args:
        config: session configuration dictionary obtained e.g. by parsing a JSON file. Note that the data contained
            in this dictionary will be modified in-place.
        cfg_ver_str: string representing the version for which the configuration was created (e.g. "0.2.0").

    Returns:
        An updated configuration dictionary that should be compatible with the current version of the framework.
    """
    if not isinstance(config, dict):
        raise AssertionError("unexpected config type")
    if not isinstance(cfg_ver_str, str) or len(cfg_ver_str.split(".")) != 3:
        raise AssertionError("unexpected checkpoint version formatting")
    from thelper import __version__ as curr_ver
    curr_ver = [int(num) for num in curr_ver.split(".")]
    cfg_ver = [int(num) for num in cfg_ver_str.split(".")]
    if (cfg_ver[0] > curr_ver[0] or (cfg_ver[0] == curr_ver[0] and cfg_ver[1] > curr_ver[1]) or
       (cfg_ver[0:2] == curr_ver[0:2] and cfg_ver[2] > curr_ver[2])):
        raise AssertionError("cannot migrate configs from future versions!")
    if cfg_ver == [0, 0, 0]:
        logger.warning("trying to migrate config from v0.0.0; all bets are off")
    else:
        logger.info("trying to migrate config from v%s" % cfg_ver_str)
    if cfg_ver[0] <= 0 and cfg_ver[1] < 1:
        # must search for name-value parameter lists and convert them to dictionaries
        def name_value_replacer(cfg):
            if isinstance(cfg, dict):
                for key, val in cfg.items():
                    if (key == "params" or key == "parameters") and isinstance(val, list) and \
                       all([isinstance(p, dict) and list(p.keys()) == ["name", "value"] for p in val]):
                        cfg["params"] = {param["name"]: name_value_replacer(param["value"]) for param in val}
                        if key == "parameters":
                            del cfg["parameters"]
                    elif isinstance(val, (dict, list)):
                        cfg[key] = name_value_replacer(val)
            elif isinstance(cfg, list):
                for idx, val in enumerate(cfg):
                    cfg[idx] = name_value_replacer(val)
            return cfg
        config = name_value_replacer(config)
        # must replace "data_config" section by "loaders"
        if "data_config" in config:
            config["loaders"] = config["data_config"]
            del config["data_config"]
        # remove deprecated name attribute for models
        if "model" in config and isinstance(config["model"], dict) and "name" in config["model"]:
            del config["model"]["name"]
        # must update import targets wrt class name refactorings
        def import_refactoring(cfg):  # noqa: E306
            if isinstance(cfg, dict):
                for key, val in cfg.items():
                    cfg[key] = import_refactoring(val)
            elif isinstance(cfg, list):
                for idx, val in enumerate(cfg):
                    cfg[idx] = import_refactoring(val)
            elif isinstance(cfg, str) and cfg.startswith("thelper."):
                cfg = thelper.utils.resolve_import(cfg)
            return cfg
        config = import_refactoring(config)
        if "trainer" in config and isinstance(config["trainer"], dict):
            trainer_cfg = config["trainer"]
            # move 'loss' section to 'optimization' section
            if "loss" in trainer_cfg:
                if "optimization" not in trainer_cfg or not isinstance(trainer_cfg["optimization"], dict):
                    trainer_cfg["optimization"] = {}
                trainer_cfg["optimization"]["loss"] = trainer_cfg["loss"]
                del trainer_cfg["loss"]
            # replace all devices with cuda:all
            if "train_device" in trainer_cfg:
                del trainer_cfg["train_device"]
            if "valid_device" in trainer_cfg:
                del trainer_cfg["valid_device"]
            if "test_device" in trainer_cfg:
                del trainer_cfg["test_device"]
            if "device" not in trainer_cfg:
                trainer_cfg["device"] = "cuda:all"
            # remove params from trainer config
            if "params" in trainer_cfg:
                if not isinstance(trainer_cfg["params"], (dict, list)) or trainer_cfg["params"]:
                    logger.warning("removing non-empty parameter section from trainer config")
                del trainer_cfg["params"]
        cfg_ver = [0, 1, 0]  # set ver for next update step
    if cfg_ver[0] <= 0 and cfg_ver[1] <= 1:
        # remove 'force_convert' flags from all transform pipelines + build augment pipeline wrappers
        def remove_force_convert(cfg):  # noqa: E306
            if isinstance(cfg, list):
                for idx, stage in enumerate(cfg):
                    cfg[idx] = remove_force_convert(stage)
            elif isinstance(cfg, dict):
                if "parameters" in cfg:
                    cfg["params"] = cfg["parameters"]
                    del cfg["parameters"]
                if "operation" in cfg and cfg["operation"] == "thelper.transforms.TransformWrapper":
                    if "params" in cfg and "force_convert" in cfg["params"]:
                        del cfg["params"]["force_convert"]
                for key, stage in cfg.items():
                    cfg[key] = remove_force_convert(stage)
            return cfg
        for pipeline in ["base_transforms", "train_augments", "valid_augments", "test_augments"]:
            if "loaders" in config and isinstance(config["loaders"], dict) and pipeline in config["loaders"]:
                if pipeline.endswith("_augments"):
                    stages = config["loaders"][pipeline]
                    for stage in stages:
                        if "append" in stage:
                            if stage["append"]:
                                logger.warning("overriding augmentation stage ordering")
                            del stage["append"]
                        if "operation" in stage and stage["operation"] == "Augmentor.Pipeline":
                            if "params" in stage:
                                stage["params"] = stage["params"]["operations"]
                            elif "parameters" in stage:
                                stage["params"] = stage["parameters"]["operations"]
                                del stage["parameters"]
                    config["loaders"][pipeline] = {"append": False, "transforms": remove_force_convert(stages)}
                else:
                    config["loaders"][pipeline] = remove_force_convert(config["loaders"][pipeline])
        cfg_ver = [0, 2, 0]  # set ver for next update step
    if cfg_ver[0] <= 0 and cfg_ver[1] <= 2 and cfg_ver[2] < 5:
        # TODO: add scheduler 0-based step fix here? (unlikely to cause serious issues)
        # cfg_ver = [0, 2, 5]  # set ver for next update step
        pass
    # if cfg_ver[0] <= x and cfg_ver[1] <= y and cfg_ver[2] <= z:
    #     ... add more compatibility fixes here
    return config


def download_file(url, root, filename, md5=None):
    """Downloads a file from a given URL to a local destination.

    Args:
        url: path to query for the file (query will be based on urllib).
        root: destination folder where the file should be saved.
        filename: destination name for the file.
        md5: optional, for md5 integrity check.

    Returns:
        The path to the downloaded file.
    """
    # inspired from torchvision.datasets.utils.download_url; no dep check
    from six.moves import urllib
    root = os.path.expanduser(root)
    fpath = os.path.join(root, filename)
    try:
        os.makedirs(root)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise
    if not os.path.isfile(fpath):
        logger.info("Downloading %s to %s ..." % (url, fpath))
        urllib.request.urlretrieve(url, fpath, reporthook)
        sys.stdout.write("\r")
        sys.stdout.flush()
    if md5 is not None:
        import hashlib
        md5o = hashlib.md5()
        with open(fpath, 'rb') as f:
            for chunk in iter(lambda: f.read(1024 * 1024), b''):
                md5o.update(chunk)
        md5c = md5o.hexdigest()
        if md5c != md5:
            raise AssertionError("md5 check failed for '%s'" % fpath)
    return fpath


def extract_tar(filepath, root, flags="r:gz"):
    """Extracts the content of a tar file to a specific location.

    Args:
        filepath: location of the tar archive.
        root: where to extract the archive's content.
        flags: extra flags passed to ``tarfile.open``.
    """
    import tarfile

    class _FileWrapper(io.FileIO):
        def __init__(self, path, *args, **kwargs):
            self.start_time = time.time()
            self._size = os.path.getsize(path)
            super().__init__(path, *args, **kwargs)

        def read(self, *args, **kwargs):
            duration = time.time() - self.start_time
            progress_size = self.tell()
            speed = str(int(progress_size / (1024 * duration))) if duration > 0 else "?"
            percent = min(int(progress_size * 100 / self._size), 100)
            sys.stdout.write("\r\t=> extracted %d%% (%d MB) @ %s KB/s..." %
                             (percent, progress_size / (1024 * 1024), speed))
            sys.stdout.flush()
            return io.FileIO.read(self, *args, **kwargs)

    cwd = os.getcwd()
    tar = tarfile.open(fileobj=_FileWrapper(filepath), mode=flags)
    os.chdir(root)
    tar.extractall()
    tar.close()
    os.chdir(cwd)
    sys.stdout.write("\r")
    sys.stdout.flush()


def reporthook(count, block_size, total_size):
    """Report hook used to display a download progression bar when using urllib requests."""
    global start_time
    if count == 0:
        start_time = time.time()
        return
    duration = time.time() - start_time
    progress_size = int(count * block_size)
    speed = str(int(progress_size / (1024 * duration))) if duration > 0 else "?"
    percent = min(int(count * block_size * 100 / total_size), 100)
    sys.stdout.write("\r\t=> downloaded %d%% (%d MB) @ %s KB/s..." %
                     (percent, progress_size / (1024 * 1024), speed))
    sys.stdout.flush()


def resolve_import(fullname):
    # type: (AnyStr) -> AnyStr
    """
    Class name resolver.

    Takes a string corresponding to a module and class fullname to be imported with ``thelper.utils.import_class``
    and resolves any back compatibility issues related to renamed or moved classes.

    Args:
        fullname: the fully qualified class name to be resolved.

    Returns:
        The resolved class fullname.
    """
    cases = [
        ('thelper.modules', 'thelper.nn'),
        ('thelper.transforms.ImageTransformWrapper', 'thelper.transforms.TransformWrapper'),
        ('thelper.transforms.wrappers.ImageTransformWrapper', 'thelper.transforms.wrappers.TransformWrapper'),
    ]
    old_name = fullname
    for old, new in cases:
        fullname = fullname.replace(old, new)
    if old_name != fullname:
        logger.warning("class fullname '{!s}' was resolved to '{!s}'.".format(old_name, fullname))
    return fullname


def import_class(fullname):
    """General-purpose runtime class importer.

    Args:
        fullname: the fully qualified class name to be imported.

    Returns:
        The imported class.
    """
    fullname = resolve_import(fullname)
    module_name, class_name = fullname.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def import_function(fullname, params=None):
    """General-purpose runtime function importer, with support for param binding.

    Args:
        fullname: the fully qualified function name to be imported.
        params: optional params dictionary to bind to the function call via functools.

    Returns:
        The imported function, with optionally bound parameters.
    """
    func = import_class(fullname)
    if params is not None:
        if not isinstance(params, dict):
            raise AssertionError("unexpected params dict type")
        return functools.partial(func, **params)
    return func


def check_func_signature(func,      # type: Callable
                         params     # type: List[AnyStr]
                         ):         # type: (...) -> None
    """
    Checks whether the signature of a function matches the expected parameter list.

    .. seealso::
        | :class:`thelper.typedefs.IterCallbackType`
        | :class:`thelper.typedefs.IterCallbackParams`
    """
    if func is None or not callable(func):
        raise AssertionError("invalid function object")
    if params is not None:
        if not isinstance(params, list) or not all([isinstance(p, str) for p in params]):
            raise AssertionError("unexpected param name list format")
        import inspect
        func_sig = inspect.signature(func)
        for p in params:
            if p not in func_sig.parameters:
                raise AssertionError("function missing parameter '%s'" % p)


def get_class_logger(skip=0):
    """Shorthand to get logger for current class frame."""
    return logging.getLogger(get_caller_name(skip + 1).rsplit(".", 1)[0])


def get_func_logger(skip=0):
    """Shorthand to get logger for current function frame."""
    return logging.getLogger(get_caller_name(skip + 1))


def get_caller_name(skip=2):
    # source: https://gist.github.com/techtonik/2151727
    """Returns the name of a caller in the format module.class.method.

    Args:
        skip: specifies how many levels of stack to skip while getting the caller.

    Returns:
        An empty string is returned if skipped levels exceed stack height; otherwise,
        returns the requested caller name.
    """

    def stack_(frame):
        frame_list = []
        while frame:
            frame_list.append(frame)
            frame = frame.f_back
        return frame_list

    # noinspection PyProtectedMember
    stack = stack_(sys._getframe(1))
    start = 0 + skip
    if len(stack) < start + 1:
        return ""
    parent_frame = stack[start]
    name = []
    module = inspect.getmodule(parent_frame)
    # `modname` can be None when frame is executed directly in console
    if module:
        name.append(module.__name__)
    # detect class name
    if "self" in parent_frame.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parent_frame.f_locals["self"].__class__.__name__)
    codename = parent_frame.f_code.co_name
    if codename != "<module>":  # top level usually
        name.append(codename)  # function or a method
    del parent_frame
    return ".".join(name)


def get_key(key, config, msg=None):
    """Returns a value given a dictionary key, throwing if not available."""
    if isinstance(key, list):
        if len(key) <= 1:
            if msg is not None:
                raise AssertionError(msg)
            else:
                raise AssertionError("must provide at least two valid keys to test")
        for k in key:
            if k in config:
                return config[k]
        if msg is not None:
            raise AssertionError(msg)
        else:
            raise AssertionError("config dictionary missing a field named as one of '%s'" % str(key))
    else:
        if key not in config:
            if msg is not None:
                raise AssertionError(msg)
            else:
                raise AssertionError("config dictionary missing '%s' field" % key)
        else:
            return config[key]


def get_key_def(key, config, default=None, msg=None):
    """Returns a value given a dictionary key, or the default value if it cannot be found."""
    if isinstance(key, list):
        if len(key) <= 1:
            if msg is not None:
                raise AssertionError(msg)
            else:
                raise AssertionError("must provide at least two valid keys to test")
        for k in key:
            if k in config:
                return config[k]
        return default
    else:
        if key not in config:
            return default
        else:
            return config[key]


def get_log_stamp():
    """Returns a print-friendly and filename-friendly identification string containing platform and time."""
    return str(platform.node()) + "-" + time.strftime("%Y%m%d-%H%M%S")


def get_git_stamp():
    """Returns a print-friendly SHA signature for the framework's underlying git repository (if found)."""
    try:
        import git
        try:
            repo = git.Repo(path=os.path.abspath(__file__), search_parent_directories=True)
            sha = repo.head.object.hexsha
            return str(sha)
        except (AttributeError, git.InvalidGitRepositoryError):
            return "unknown"
    except (ImportError, AttributeError):
        return "unknown"


def get_env_list():
    """Returns a list of all packages installed in the current environment.

    If the required packages cannot be imported, the returned list will be empty. Note that some
    packages may not be properly detected by this approach, and it is pretty hacky, so use it with
    a grain of salt (i.e. logging is fine).
    """
    try:
        import pip
        # noinspection PyUnresolvedReferences
        pkgs = pip.get_installed_distributions()
        return sorted(["%s %s" % (pkg.key, pkg.version) for pkg in pkgs])
    except (ImportError, AttributeError):
        try:
            import pkg_resources as pkgr
            return sorted([str(pkg) for pkg in pkgr.working_set])
        except (ImportError, AttributeError):
            return []


def str2size(input_str):
    """Returns a (WIDTH, HEIGHT) integer size tuple from a string formatted as 'WxH'."""
    if not isinstance(input_str, str):
        raise AssertionError("unexpected input type")
    display_size_str = input_str.split('x')
    if len(display_size_str) != 2:
        raise AssertionError("bad size string formatting")
    return tuple([max(int(substr), 1) for substr in display_size_str])


def str2bool(s):
    """Converts a string to a boolean.

    If the lower case version of the provided string matches any of 'true', '1', or
    'yes', then the function returns ``True``.
    """
    if isinstance(s, bool):
        return s
    if isinstance(s, (int, float)):
        return s != 0
    if isinstance(s, str):
        positive_flags = ["true", "1", "yes"]
        return s.lower() in positive_flags
    raise AssertionError("unrecognized input type")


def clipstr(s, size, fill=" "):
    """Clips a string to a specific length, with an optional fill character."""
    if len(s) > size:
        s = s[:size]
    if len(s) < size:
        s = fill * (size - len(s)) + s
    return s


def lreplace(string, old_prefix, new_prefix):
    """Replaces a single occurrence of `old_prefix` in the given string by `new_prefix`."""
    return re.sub(r'^(?:%s)+' % re.escape(old_prefix), lambda m: new_prefix * (m.end() // len(old_prefix)), string)


def query_yes_no(question, default=None, bypass=None):
    """Asks the user a yes/no question and returns the answer.

    Args:
        question: the string that is presented to the user.
        default: the presumed answer if the user just hits ``<Enter>``. It must be 'yes',
            'no', or ``None`` (meaning an answer is required).
        bypass: the option to select if the ``bypass_queries`` global variable is set to
            ``True``. Can be ``None``, in which case the function will throw an exception.

    Returns:
        ``True`` for 'yes', or ``False`` for 'no' (or their respective variations).
    """
    valid = {"yes": True, "ye": True, "y": True, "no": False, "n": False}
    if bypass is not None and (not isinstance(bypass, str) or bypass not in valid):
        raise AssertionError("unexpected bypass value")
    if bypass_queries:
        if bypass is None:
            raise AssertionError("cannot bypass interactive query, no default value provided")
        return valid[bypass]
    if (isinstance(default, bool) and default) or \
       (isinstance(default, str) and default.lower() in ["yes", "ye", "y"]):
        prompt = " [Y/n] "
    elif (isinstance(default, bool) and not default) or \
         (isinstance(default, str) and default.lower() in ["no", "n"]):
        prompt = " [y/N] "
    else:
        prompt = " [y/n] "
    sys.stdout.flush()
    sys.stderr.flush()
    time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
    while True:
        sys.stdout.write(question + prompt + "\n>> ")
        choice = input().lower()
        if default is not None and choice == "":
            if isinstance(default, str):
                return valid[default]
            else:
                return default
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes/y' or 'no/n'.\n")


def query_string(question, choices=None, default=None, allow_empty=False, bypass=None):
    """Asks the user a question and returns the answer (a generic string).

    Args:
        question: the string that is presented to the user.
        choices: a list of predefined choices that the user can pick from. If
            ``None``, then whatever the user types will be accepted.
        default: the presumed answer if the user just hits ``<Enter>``. If ``None``,
            then an answer is required to continue.
        allow_empty: defines whether an empty answer should be accepted.
        bypass: the returned value if the ``bypass_queries`` global variable is set to
            ``True``. Can be ``None``, in which case the function will throw an exception.

    Returns:
        The string entered by the user.
    """
    if bypass_queries:
        if bypass is None:
            raise AssertionError("cannot bypass interactive query, no default value provided")
        return bypass
    sys.stdout.flush()
    sys.stderr.flush()
    time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
    while True:
        msg = question
        if choices is not None:
            msg += "\n\t(choices=%s)" % str(choices)
        if default is not None:
            msg += "\n\t(default=%s)" % default
        sys.stdout.write(msg + "\n>> ")
        answer = input()
        if answer == "":
            if default is not None:
                return default
            elif allow_empty:
                return answer
        elif choices is not None:
            if answer in choices:
                return answer
        else:
            return answer
        sys.stdout.write("Please respond with a valid string.\n")


def get_save_dir(out_root, dir_name, config=None, resume=False):
    """Returns a directory path in which the app can save its data.

    If a folder with name ``dir_name`` already exists in the directory ``out_root``, then the user will be
    asked to pick a new name. If the user refuses, ``sys.exit(1)`` is called. If config is not ``None``, it
    will be saved to the output directory as a json file. Finally, a ``logs`` directory will also be created
    in the output directory for writing logger files.

    Args:
        out_root: path to the directory root where the save directory should be created.
        dir_name: name of the save directory to create. If it already exists, a new one will be requested.
        config: dictionary of app configuration parameters. Used to overwrite i/o queries, and will be
            written to the save directory in json format to test writing. Default is ``None``.
        resume: specifies whether this session is new, or resumed from an older one (in the latter
            case, overwriting is allowed, and the user will never have to choose a new folder)

    Returns:
        The path to the created save directory for this session.
    """
    func_logger = get_func_logger()
    save_dir = out_root
    if save_dir is None:
        time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
        save_dir = query_string("Please provide the path to where session directories should be created/saved:")
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    save_dir = os.path.join(save_dir, dir_name)
    if not resume:
        overwrite = str2bool(config["overwrite"]) if config is not None and "overwrite" in config else False
        time.sleep(0.25)  # to make sure all debug/info prints are done, and we see the question
        while os.path.exists(save_dir) and not overwrite:
            abs_save_dir = os.path.abspath(save_dir).replace("\\", "/")
            overwrite = query_yes_no("Training session at '%s' already exists; overwrite?" % abs_save_dir, bypass="y")
            if not overwrite:
                save_dir = query_string("Please provide a new save directory path:")
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        if config is not None:
            config_backup_path = os.path.join(save_dir, "config.latest.json")
            with open(config_backup_path, "w") as fd:
                json.dump(config, fd, indent=4, sort_keys=False)
    else:
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        if config is not None:
            config_backup_path = os.path.join(save_dir, "config.latest.json")
            if os.path.exists(config_backup_path):
                config_backup = json.load(open(config_backup_path, "r"))
                if config_backup != config:
                    query_msg = "Config backup in '%s' differs from config loaded through checkpoint; overwrite?" \
                                % config_backup_path
                    answer = query_yes_no(query_msg, bypass="y")
                    if answer:
                        func_logger.warning("config mismatch with previous run; "
                                            "will overwrite latest backup in save directory")
                    else:
                        func_logger.error("config mismatch with previous run; user aborted")
                        sys.exit(1)
            with open(config_backup_path, "w") as fd:
                json.dump(config, fd, indent=4, sort_keys=False)
    logs_dir = os.path.join(save_dir, "logs")
    if not os.path.exists(logs_dir):
        os.mkdir(logs_dir)
    return save_dir


def safe_crop(image, tl, br, bordertype=cv.BORDER_CONSTANT, borderval=0):
    """Safely crops a region from within an image, padding borders if needed.

    Args:
        image: the image to crop (provided as a numpy array).
        tl: a tuple or list specifying the (x,y) coordinates of the top-left crop corner.
        br: a tuple or list specifying the (x,y) coordinates of the bottom-right crop corner.
        bordertype: border copy type to use when the image is too small for the required crop size.
            See ``cv2.copyMakeBorder`` for more information.
        borderval: border value to use when the image is too small for the required crop size. See
            ``cv2.copyMakeBorder`` for more information.

    Returns:
        The cropped image.
    """
    if not isinstance(image, np.ndarray):
        raise AssertionError("expected input image to be numpy array")
    if isinstance(tl, tuple):
        tl = list(tl)
    if isinstance(br, tuple):
        br = list(br)
    if not isinstance(tl, list) or not isinstance(br, list):
        raise AssertionError("expected tl/br coords to be provided as tuple or list")
    if tl[0] < 0 or tl[1] < 0 or br[0] > image.shape[1] or br[1] > image.shape[0]:
        image = cv.copyMakeBorder(image, max(-tl[1], 0), max(br[1] - image.shape[0], 0),
                                  max(-tl[0], 0), max(br[0] - image.shape[1], 0),
                                  borderType=bordertype, value=borderval)
        if tl[0] < 0:
            br[0] -= tl[0]
            tl[0] = 0
        if tl[1] < 0:
            br[1] -= tl[1]
            tl[1] = 0
    return image[tl[1]:br[1], tl[0]:br[0], ...]


def get_bgr_from_hsl(hue, sat, light):
    """Converts a single HSL triplet (0-360 hue, 0-1 sat & lightness) into an 8-bit RGB triplet."""
    # this function is not intended for fast conversions; use OpenCV's cvtColor for large-scale stuff
    if hue < 0 or hue > 360:
        raise AssertionError("invalid hue")
    if sat < 0 or sat > 1:
        raise AssertionError("invalid saturation")
    if light < 0 or light > 1:
        raise AssertionError("invalid lightness")
    if sat == 0:
        return (int(np.clip(round(light * 255), 0, 255)),) * 3
    if light == 0:
        return 0, 0, 0
    if light == 1:
        return 255, 255, 255

    def h2rgb(_p, _q, _t):
        if _t < 0:
            _t += 1
        if _t > 1:
            _t -= 1
        if _t < 1 / 6:
            return _p + (_q - _p) * 6 * _t
        if _t < 1 / 2:
            return _q
        if _t < 2 / 3:
            return _p + (_q - _p) * (2 / 3 - _t) * 6
        return _p

    q = light * (1 + sat) if (light < 0.5) else light + sat - light*sat
    p = 2 * light - q
    h = hue / 360
    return (int(np.clip(round(h2rgb(p, q, h - 1 / 3) * 255), 0, 255)),
            int(np.clip(round(h2rgb(p, q, h) * 255), 0, 255)),
            int(np.clip(round(h2rgb(p, q, h + 1 / 3) * 255), 0, 255)))


def get_displayable_image(image,                # type: thelper.typedefs.ArrayType
                          grayscale=False,      # type: Optional[bool]
                          ):                    # type: (...) -> thelper.typedefs.ArrayType
    """Returns a 'displayable' image that has been normalized and padded to three channels."""
    if image.ndim != 3:
        raise AssertionError("indexing should return a pre-squeezed array")
    if image.shape[2] == 2:
        image = np.dstack((image, image[:, :, 0]))
    elif image.shape[2] > 3:
        image = image[..., :3]
    if grayscale and image.shape[2] != 1:
        image = cv.cvtColor(image, cv.COLOR_BGR2GRAY)
    image_normalized = np.empty_like(image, dtype=np.uint8).copy()  # copy needed here due to ocv 3.3 bug
    cv.normalize(image, image_normalized, 0, 255, cv.NORM_MINMAX, dtype=cv.CV_8U)
    return image_normalized


def get_displayable_heatmap(array,              # type: thelper.typedefs.ArrayType
                            convert_rgb=True,   # type: Optional[bool]
                            ):                  # type: (...) -> thelper.typedefs.ArrayType
    """Returns a 'displayable' array that has been min-maxed and mapped to color triplets."""
    if array.ndim != 2:
        array = np.squeeze(array)
    if array.ndim != 2:
        raise AssertionError("indexing should return a pre-squeezed array")
    array_normalized = np.empty_like(array, dtype=np.uint8).copy()  # copy needed here due to ocv 3.3 bug
    cv.normalize(array, array_normalized, 0, 255, cv.NORM_MINMAX, dtype=cv.CV_8U)
    heatmap = cv.applyColorMap(array_normalized, cv.COLORMAP_JET)
    if convert_rgb:
        heatmap = cv.cvtColor(heatmap, cv.COLOR_BGR2RGB)
    return heatmap


def draw_histogram(data,                # type: thelper.typedefs.ArrayType
                   bins=50,             # type: Optional[int]
                   xlabel="",           # type: Optional[thelper.typedefs.LabelType]
                   ylabel="Proportion"  # type: Optional[thelper.typedefs.LabelType]
                   ):                   # type: (...) -> thelper.typedefs.DrawingType
    """Draws and returns a histogram figure using pyplot."""
    fig, ax = plt.subplots()
    ax.hist(data, density=True, bins=bins)
    if len(ylabel) > 0:
        ax.set_ylabel(ylabel)
    if len(xlabel) > 0:
        ax.set_xlabel(xlabel)
    ax.set_xlim(xmin=0)
    fig.show()
    return fig


def draw_popbars(labels,                # type: thelper.typedefs.LabelList
                 counts,                # type: int
                 xlabel="",             # type: Optional[thelper.typedefs.LabelType]
                 ylabel="Pop. Count",   # type: Optional[thelper.typedefs.LabelType]
                 ):                     # type: (...) -> thelper.typedefs.DrawingType
    """Draws and returns a bar histogram figure using pyplot."""
    fig, ax = plt.subplots()
    xrange = range(len(labels))
    ax.bar(xrange, counts, align="center")
    if len(ylabel) > 0:
        ax.set_ylabel(ylabel)
    if len(xlabel) > 0:
        ax.set_xlabel(xlabel)
    ax.set_xticks(xrange)
    ax.set_xticklabels(labels)
    ax.tick_params(axis="x", labelsize="8", labelrotation=45)
    fig.show()
    return fig


def draw_classifs(images,               # type: thelper.typedefs.OneOrManyArrayType
                  labels_gt=None,       # type: Optional[thelper.typedefs.LabelList]
                  labels_pred=None,     # type: Optional[thelper.typedefs.LabelList]
                  labels_map=None,      # type: Optional[thelper.typedefs.LabelDict]
                  redraw=None,          # type: Optional[thelper.typedefs.DrawingType]
                  use_cv2=True,         # type: Optional[bool]
                  img_shape=None,       # type: Optional[thelper.typedefs.ArrayShapeType]
                  max_img_size=None,    # type: Optional[thelper.typedefs.ArrayShapeType]
                  ):                    # type: (...) -> thelper.typedefs.DrawingType
    """Draws and returns a figure of classification results using pyplot."""
    nb_imgs = len(images) if isinstance(images, list) else images.shape[0]
    if nb_imgs < 1:
        return None
    if max_img_size is None:
        max_img_size = (800, 1600)
    grid_size_x = int(math.ceil(math.sqrt(nb_imgs)))
    grid_size_y = int(math.ceil(nb_imgs / grid_size_x))
    if grid_size_x * grid_size_y < nb_imgs:
        raise AssertionError("bad gridding for subplots")
    if use_cv2:
        img_grid_shape, img_grid = None, (None if redraw is None else redraw[1])
        for img_idx in range(nb_imgs):
            image = images[img_idx] if isinstance(images, list) else images[img_idx, ...]
            if img_shape is None:
                img_shape = image.shape
            if img_grid_shape is None:
                img_grid_shape = (img_shape[0] * grid_size_y, img_shape[1] * grid_size_x, img_shape[2])
            if img_grid is None or img_grid.shape != img_grid_shape:
                img_grid = np.zeros(img_grid_shape, dtype=np.uint8)
            if labels_gt is not None or labels_pred is not None:
                if labels_gt is not None:
                    if labels_map and labels_gt[img_idx] in labels_map:
                        curr_label_gt = labels_map[labels_gt[img_idx]]
                    else:
                        curr_label_gt = labels_gt[img_idx]
                else:
                    curr_label_gt = "<unknown>"
                if labels_pred is not None:
                    if labels_map and labels_pred[img_idx] in labels_map:
                        curr_label_pred = labels_map[labels_pred[img_idx]]
                    else:
                        curr_label_pred = labels_pred[img_idx]
                    xlabel = "GT={0}\nPred={1}".format(curr_label_gt, curr_label_pred)
                else:
                    xlabel = "GT={0}".format(curr_label_gt)
                image = image.copy()
                color = (255, 255, 255)
                bottom_left = (10, 40)
                image = cv.putText(image, xlabel, bottom_left, cv.FONT_HERSHEY_SIMPLEX, 0.40, color, 1, cv.LINE_AA)
            offsets = (img_idx // grid_size_x) * img_shape[0], (img_idx % grid_size_x) * img_shape[1]
            np.copyto(img_grid[offsets[0]:(offsets[0] + img_shape[0]),
                      offsets[1]:(offsets[1] + img_shape[1]), :], image)
        win_name = "classifs" if redraw is None else redraw[0]
        if img_grid is not None:
            display = img_grid[..., ::-1]
            if display.shape[0] > max_img_size[0] or display.shape[1] > max_img_size[1]:
                if display.shape[0] / max_img_size[0] > display.shape[1] / max_img_size[1]:
                    dsize = (max_img_size[0], int(round(display.shape[1] / (display.shape[0] / max_img_size[0]))))
                else:
                    dsize = (int(round(display.shape[0] / (display.shape[1] / max_img_size[1]))), max_img_size[1])
                display = cv.resize(display, (dsize[1], dsize[0]))
            cv.imshow(win_name, display)
        return win_name, img_grid
    else:
        fig, axes = redraw if redraw is not None else plt.subplots(grid_size_y, grid_size_x)
        plt.tight_layout()
        if nb_imgs == 1:
            axes = np.array(axes)
        for ax_idx, ax in enumerate(axes.reshape(-1)):
            if ax_idx < nb_imgs:
                if isinstance(images, list):
                    ax.imshow(images[ax_idx], interpolation='nearest')
                else:
                    ax.imshow(images[ax_idx, ...], interpolation='nearest')
                if labels_gt is not None or labels_pred is not None:
                    if labels_gt is not None:
                        if labels_map and labels_gt[ax_idx] in labels_map:
                            curr_label_gt = labels_map[labels_gt[ax_idx]]
                        else:
                            curr_label_gt = labels_gt[ax_idx]
                    else:
                        curr_label_gt = "<unknown>"
                    if labels_pred is not None:
                        if labels_map and labels_pred[ax_idx] in labels_map:
                            curr_label_pred = labels_map[labels_pred[ax_idx]]
                        else:
                            curr_label_pred = labels_pred[ax_idx]
                        xlabel = "GT={0}\nPred={1}".format(curr_label_gt, curr_label_pred)
                    else:
                        xlabel = "GT={0}".format(curr_label_gt)
                    ax.set_xlabel(xlabel)
            ax.set_xticks([])
            ax.set_yticks([])
        fig.show()
        return fig, axes


def draw_segments(images,                 # type: thelper.typedefs.OneOrManyArrayType
                  masks_gt,               # type: Optional[thelper.typedefs.OneOrManyArrayType]
                  masks_pred=None,        # type: Optional[thelper.typedefs.OneOrManyArrayType]
                  labels_color_map=None,  # type: Optional[thelper.typedefs.LabelColorMapType]
                  redraw=None,            # type: Optional[thelper.typedefs.DrawingType]
                  use_cv2=True,           # type: Optional[bool]
                  img_shape=None,         # type: Optional[thelper.typedefs.ArrayShapeType]
                  max_img_size=None,      # type: Optional[thelper.typedefs.ArrayShapeType]
                  ):                      # type: (...) -> thelper.typedefs.DrawingType
    """Draws and returns a figure of segmentation results using pyplot."""
    # todo: display gt if available? (currently skipped)
    nb_imgs = len(images) if isinstance(images, list) else images.shape[0]
    if nb_imgs < 1:
        return None
    if max_img_size is None:
        max_img_size = (800, 1600)
    grid_size_x = int(math.ceil(math.sqrt(nb_imgs)))
    grid_size_y = int(math.ceil(nb_imgs / grid_size_x))
    if grid_size_x * grid_size_y < nb_imgs:
        raise AssertionError("bad gridding for subplots")
    if labels_color_map is not None and isinstance(labels_color_map, dict):
        if len(labels_color_map) > 256:
            raise AssertionError("too many indices for uint8 map")
        labels_color_map_new = np.zeros((256, 1, 3), dtype=np.uint8)
        for idx, val in labels_color_map.items():
            labels_color_map_new[idx, ...] = val
        labels_color_map = labels_color_map_new
    if use_cv2:
        img_grid_shape, img_grid = None, (None if redraw is None else redraw[1])
        for img_idx in range(nb_imgs):
            image = images[img_idx] if isinstance(images, list) else images[img_idx, ...]
            if img_shape is None:
                img_shape = image.shape
            if img_grid_shape is None:
                img_grid_shape = (img_shape[0] * grid_size_y, img_shape[1] * grid_size_x, 3)
            if img_grid is None or img_grid.shape != img_grid_shape:
                img_grid = np.zeros(img_grid_shape, dtype=np.uint8)
            mask = None
            if masks_pred is not None:
                mask = masks_pred[img_idx] if isinstance(masks_pred, list) else masks_pred[img_idx, ...]
            elif masks_gt is not None:
                mask = masks_gt[img_idx] if isinstance(masks_gt, list) else masks_gt[img_idx, ...]
            if isinstance(mask, np.ndarray):
                if labels_color_map is not None:
                    mask = apply_color_map(mask, labels_color_map)
                if image.ndim == 2 or image.shape[2] != 3:
                    image = cv.cvtColor(image, cv.COLOR_GRAY2BGR)
                image = cv.addWeighted(image, 0.5, mask, 0.5, 0)
            offsets = (img_idx // grid_size_x) * img_shape[0], (img_idx % grid_size_x) * img_shape[1]
            if image.ndim < 3 or image.shape[2] == 1:
                image = cv.cvtColor(image, cv.COLOR_GRAY2BGR)
            np.copyto(img_grid[offsets[0]:(offsets[0] + img_shape[0]),
                      offsets[1]:(offsets[1] + img_shape[1]), :], image)
        win_name = "segments" if redraw is None else redraw[0]
        if img_grid is not None:
            display = img_grid[..., ::-1]
            if display.shape[0] > max_img_size[0] or display.shape[1] > max_img_size[1]:
                if display.shape[0] / max_img_size[0] > display.shape[1] / max_img_size[1]:
                    dsize = (max_img_size[0], int(round(display.shape[1] / (display.shape[0] / max_img_size[0]))))
                else:
                    dsize = (int(round(display.shape[0] / (display.shape[1] / max_img_size[1]))), max_img_size[1])
                display = cv.resize(display, (dsize[1], dsize[0]))
            cv.imshow(win_name, display)
        return win_name, img_grid
    else:
        fig, axes = redraw if redraw is not None else plt.subplots(grid_size_y, grid_size_x)
        plt.tight_layout()
        if nb_imgs == 1:
            axes = np.array(axes)
        for ax_idx, ax in enumerate(axes.reshape(-1)):
            if ax_idx < nb_imgs:
                image = images[ax_idx] if isinstance(images, list) else images[ax_idx, ...]
                if masks_pred is not None:
                    mask = masks_pred[ax_idx] if isinstance(masks_pred, list) else masks_pred[ax_idx, ...]
                    if labels_color_map is not None:
                        mask = apply_color_map(mask, labels_color_map)
                    if image.ndim == 2 or image.shape[2] != 3:
                        image = cv.cvtColor(image, cv.COLOR_GRAY2BGR)
                    image = cv.addWeighted(image, 0.5, mask, 0.5, 0)
                ax.imshow(image, interpolation='nearest')
            ax.set_xticks([])
            ax.set_yticks([])
        fig.show()
        return fig, axes


def draw_minibatch(minibatch, task, preds=None, block=False, ch_transpose=True,
                   flip_bgr=False, redraw=None, use_cv2=True):
    """Draws and returns a figure of a minibatch using pyplot or OpenCV."""
    if not isinstance(minibatch, dict):
        raise AssertionError("expected dict-based sample")
    import thelper.tasks
    if not isinstance(task, thelper.tasks.Task):
        raise AssertionError("invalid task object")
    image_key = task.get_input_key()
    if image_key is None or image_key not in minibatch:
        raise AssertionError("images not found with key '%s'" % image_key)
    images = minibatch[image_key]
    if isinstance(images, list) and all([isinstance(t, torch.Tensor) for t in images]):
        # if we have a list, it must be due to a augmentation stage
        if not all([image.shape == images[0].shape for image in images]):
            raise AssertionError("image shape mismatch throughout list")
        images = torch.cat(images, 0)  # merge all images into a single tensor
    if not isinstance(images, torch.Tensor) or images.dim() != 4:
        raise AssertionError("expected input images to be in 4-d tensor format (BxCxHxW or BxHxWxC)")
    images = images.numpy().copy()
    if ch_transpose:
        images = np.transpose(images, (0, 2, 3, 1))  # BxCxHxW to BxHxWxC
    if flip_bgr:
        images = images[..., ::-1]  # BGR to RGB
    if preds is not None:
        preds = preds.cpu()  # avoid latency for preprocessing on gpu
    image_list = [get_displayable_image(images[batch_idx, ...]) for batch_idx in range(images.shape[0])]
    if isinstance(task, thelper.tasks.Classification):
        label_key, labels = task.get_gt_key(), None
        if label_key in minibatch and minibatch[label_key] is not None:
            labels = minibatch[label_key]
            if not isinstance(labels, list) and not (isinstance(labels, torch.Tensor) and labels.dim() == 1):
                raise AssertionError("expected classification labels to be in list or 1-d tensor format")
            if isinstance(labels, list):
                if all([isinstance(l, list) for l in labels]):
                    labels = list(itertools.chain.from_iterable(labels))  # merge all augmented lists together
                if all([isinstance(t, torch.Tensor) for t in labels]):
                    labels = torch.cat(labels, 0)
            if isinstance(labels, torch.Tensor):
                labels = labels.tolist()
            if images.shape[0] != len(labels):
                raise AssertionError("images/labels count mismatch")
        if preds is not None:
            if not isinstance(preds, list) and not (isinstance(preds, torch.Tensor) and preds.dim() == 2):
                raise AssertionError("expected classification predictions to be in list or 2-d tensor format (BxC)")
            if isinstance(preds, list):
                if all([isinstance(p, list) for p in preds]):
                    preds = list(itertools.chain.from_iterable(preds))  # merge all augmented lists together
                if all([isinstance(t, torch.Tensor) for t in preds]):
                    preds = torch.cat(preds, 0)
            with torch.no_grad():
                preds = torch.squeeze(preds.topk(1, dim=1)[1], dim=1)
            if images.shape[0] != preds.shape[0]:
                raise AssertionError("images/predictions count mismatch")
            preds = preds.tolist()
        class_names_map = {idx: name for name, idx in task.get_class_idxs_map().items()}
        redraw = draw_classifs(image_list, labels_gt=labels, labels_pred=preds, labels_map=class_names_map,
                               redraw=redraw, use_cv2=use_cv2)
    elif isinstance(task, thelper.tasks.Segmentation):
        mask_key, masks = task.get_gt_key(), None
        if mask_key in minibatch and minibatch[mask_key] is not None:
            masks = minibatch[mask_key]
            if not isinstance(masks, list) and not (isinstance(masks, torch.Tensor) and masks.dim() == 3):
                raise AssertionError("expected segmentation masks to be in list or 3-d tensor format (BxHxW)")
            if isinstance(masks, list):
                if all([isinstance(m, list) for m in masks]):
                    masks = list(itertools.chain.from_iterable(masks))  # merge all augmented lists together
                masks = torch.cat(masks, 0)  # merge all masks into a single tensor
            if images.shape[0:3] != masks.shape:
                raise AssertionError("images/masks shape mismatch")
            masks = masks.numpy().copy()
        if preds is not None:
            if not isinstance(preds, list) and not (isinstance(preds, torch.Tensor) and preds.dim() == 4):
                raise AssertionError("expected segmentation preds to be in list or 3-d tensor format (BxCxHxW)")
            if isinstance(preds, list):
                if all([isinstance(p, list) for p in preds]):
                    preds = list(itertools.chain.from_iterable(preds))  # merge all augmented lists together
                preds = torch.cat(preds, 0)  # merge all preds into a single tensor
            with torch.no_grad():
                preds = torch.squeeze(preds.topk(1, dim=1)[1], dim=1)
            if images.shape[0:3] != preds.shape:
                raise AssertionError("images/preds shape mismatch")
            preds = preds.numpy().copy()
        name_color_map = task.get_color_map()
        if name_color_map is not None:
            idx_color_map = {idx: name_color_map[name] for name, idx in task.get_class_idxs_map().items()}
        else:
            idx_color_map = {idx: get_label_color_mapping(idx) for idx in task.get_class_idxs_map().values()}
        redraw = draw_segments(image_list, masks_gt=masks, masks_pred=preds, labels_color_map=idx_color_map,
                               redraw=redraw, use_cv2=use_cv2)
    elif isinstance(task, thelper.tasks.Regression):
        target_key, targets = task.get_gt_key(), None
        if target_key in minibatch and minibatch[target_key] is not None:
            targets = minibatch[target_key]
            if not isinstance(targets, list) and not (isinstance(targets, torch.Tensor) and
                                                      targets.shape[0] == images.shape[0]):
                raise AssertionError("expected targets to be in list or tensor format (Bx...)")
            if isinstance(targets, list):
                if all([isinstance(t, list) for t in targets]):
                    targets = list(itertools.chain.from_iterable(targets))  # merge all augmented lists together
                targets = torch.cat(targets, 0)  # merge all masks into a single tensor
            targets = targets.numpy().copy()
        if preds is not None:
            if not isinstance(preds, list) and not (isinstance(preds, torch.Tensor) and targets.shape == preds.shape):
                raise AssertionError("expected preds to be in list or target-compat tensor shape (Bx...)")
            if isinstance(preds, list):
                if all([isinstance(p, list) for p in preds]):
                    preds = list(itertools.chain.from_iterable(preds))  # merge all augmented lists together
                preds = torch.cat(preds, 0)  # merge all preds into a single tensor
            if targets is not None and preds.shape != targets.shape:
                raise AssertionError("preds/targets shape mismatch")
            preds = preds.numpy().copy()
        if targets is not None:
            if ((targets.ndim == 4 and targets.shape[1] == 1) or targets.ndim == 3) \
                    and targets.shape[-2:] == images.shape[1:3]:
                image_list = [get_displayable_image(images[batch_idx, ...], grayscale=True)
                              for batch_idx in range(images.shape[0])]
                tgt_heatmap_list = [get_displayable_heatmap(targets[batch_idx, ...])
                                    for batch_idx in range(images.shape[0])]
                pred_heatmap_list = None
                if preds is not None:
                    pred_heatmap_list = [get_displayable_heatmap(preds[batch_idx, ...])
                                         for batch_idx in range(images.shape[0])]
                redraw = draw_segments(image_list, masks_gt=tgt_heatmap_list, masks_pred=pred_heatmap_list,
                                       redraw=redraw, use_cv2=use_cv2)
            elif targets.ndim == 1 and targets.shape[0] == images.shape[0]:
                image_list = [get_displayable_image(images[batch_idx, ...]) for batch_idx in range(images.shape[0])]
                redraw = draw_classifs(image_list, labels_gt=targets, labels_pred=preds, redraw=redraw, use_cv2=use_cv2)
            else:
                redraw = draw_classifs(image_list, redraw=redraw, use_cv2=use_cv2)  # draw only images
        else:
            redraw = draw_classifs(image_list, redraw=redraw, use_cv2=use_cv2)  # draw only images
    else:
        raise AssertionError("unhandled drawing mode, missing impl")
    if use_cv2:
        cv.waitKey(0 if block else 1)
        return redraw
    else:
        if block:
            plt.show(block=block)
            return None
        plt.pause(0.5)
        return redraw


# noinspection PyUnusedLocal
def draw_errbars(labels,                # type: thelper.typedefs.LabelList
                 min_values,            # type: thelper.typedefs.ArrayType
                 max_values,            # type: thelper.typedefs.ArrayType
                 stddev_values,         # type: thelper.typedefs.ArrayType
                 mean_values,           # type: thelper.typedefs.ArrayType
                 xlabel="",             # type: thelper.typedefs.LabelType
                 ylabel="Raw Value"     # type: thelper.typedefs.LabelType
                 ):                     # type: (...) -> thelper.typedefs.DrawingType
    """Draws and returns an error bar histogram figure using pyplot."""
    if min_values.shape != max_values.shape \
            or min_values.shape != stddev_values.shape \
            or min_values.shape != mean_values.shape:
        raise AssertionError("input dim mismatch")
    if len(min_values.shape) != 1 and len(min_values.shape) != 2:
        raise AssertionError("input dim unexpected")
    if len(min_values.shape) == 1:
        np.expand_dims(min_values, 1)
        np.expand_dims(max_values, 1)
        np.expand_dims(stddev_values, 1)
        np.expand_dims(mean_values, 1)
    nb_subplots = min_values.shape[1]
    fig, axs = plt.subplots(nb_subplots)
    xrange = range(len(labels))
    for ax_idx in range(nb_subplots):
        ax = axs[ax_idx]
        ax.locator_params(nbins=nb_subplots)
        ax.errorbar(xrange, mean_values[:, ax_idx], stddev_values[:, ax_idx], fmt='ok', lw=3)
        ax.errorbar(xrange, mean_values[:, ax_idx], [mean_values[:, ax_idx] - min_values[:, ax_idx],
                                                     max_values[:, ax_idx] - mean_values[:, ax_idx]],
                    fmt='.k', ecolor='gray', lw=1)
        ax.set_xticks(xrange)
        ax.set_xticklabels(labels, visible=(ax_idx == nb_subplots - 1))
        ax.set_title("Band %d" % (ax_idx + 1))
        ax.tick_params(axis="x", labelsize="6", labelrotation=45)
    plt.tight_layout()
    fig.show()
    return fig


def draw_roc_curve(fpr, tpr, labels=None, size_inch=(5, 5), dpi=320):
    """Draws and returns an ROC curve figure using pyplot."""
    if not isinstance(fpr, np.ndarray) or not isinstance(tpr, np.ndarray):
        raise AssertionError("invalid inputs")
    if fpr.shape != tpr.shape:
        raise AssertionError("mismatched input sizes")
    if fpr.ndim == 1:
        fpr = np.expand_dims(fpr, 0)
    if tpr.ndim == 1:
        tpr = np.expand_dims(tpr, 0)
    if labels is not None:
        if isinstance(labels, str):
            labels = [labels]
        if len(labels) != fpr.shape[0]:
            raise AssertionError("should have one label per curve")
    else:
        labels = [None] * fpr.shape[0]
    fig = plt.figure(num="roc", figsize=size_inch, dpi=dpi, facecolor="w", edgecolor="k")
    fig.clf()
    ax = fig.add_subplot(1, 1, 1)
    for idx, label in enumerate(labels):
        auc = sklearn.metrics.auc(fpr[idx, ...], tpr[idx, ...])
        if label is not None:
            ax.plot(fpr[idx, ...], tpr[idx, ...], "b", label=("%s [auc = %0.3f]" % (label, auc)))
        else:
            ax.plot(fpr[idx, ...], tpr[idx, ...], "b", label=("auc = %0.3f" % auc))
    ax.legend(loc="lower right")
    ax.plot([0, 1], [0, 1], 'r--')
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.set_ylabel("True Positive Rate")
    ax.set_xlabel("False Positive Rate")
    fig.set_tight_layout(True)
    return fig


def draw_confmat(confmat, class_list, size_inch=(5, 5), dpi=320, normalize=False, keep_unset=False):
    """Draws and returns an a confusion matrix figure using pyplot."""
    if not isinstance(confmat, np.ndarray) or not isinstance(class_list, list):
        raise AssertionError("invalid inputs")
    if confmat.ndim != 2:
        raise AssertionError("invalid confmat shape")
    if not keep_unset and "<unset>" in class_list:
        unset_idx = class_list.index("<unset>")
        del class_list[unset_idx]
        np.delete(confmat, unset_idx, 0)
        np.delete(confmat, unset_idx, 1)
    if normalize:
        row_sums = confmat.sum(axis=1)[:, np.newaxis]
        confmat = np.nan_to_num(confmat.astype(np.float) / np.maximum(row_sums, 0.0001))
    fig = plt.figure(num="confmat", figsize=size_inch, dpi=dpi, facecolor="w", edgecolor="k")
    fig.clf()
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(confmat, cmap=plt.cm.Blues)
    labels = [clipstr(label, 9) for label in class_list]
    tick_marks = np.arange(len(labels))
    ax.set_xlabel("Predicted", fontsize=7)
    ax.set_xticks(tick_marks)
    ax.set_xticklabels(labels, fontsize=4, rotation=-90, ha="center")
    ax.xaxis.set_label_position("bottom")
    ax.xaxis.tick_bottom()
    ax.set_ylabel("Real", fontsize=7)
    ax.set_yticks(tick_marks)
    ax.set_yticklabels(labels, fontsize=4, va="center")
    ax.yaxis.set_label_position("left")
    ax.yaxis.tick_left()
    thresh = confmat.max() / 2.
    for i, j in itertools.product(range(confmat.shape[0]), range(confmat.shape[1])):
        if not normalize:
            txt = ("%d" % confmat[i, j]) if confmat[i, j] != 0 else "."
        else:
            if confmat[i, j] >= 0.01:
                txt = "%.02f" % confmat[i, j]
            else:
                txt = "~0" if confmat[i, j] > 0 else "."
        color = "white" if confmat[i, j] > thresh else "black"
        ax.text(j, i, txt, horizontalalignment="center", fontsize=4, verticalalignment="center", color=color)
    fig.set_tight_layout(True)
    return fig


def draw_bboxes(image, rects, labels=None, confidences=None, win_size=None, thickness=1, show=True):
    """Draws and returns an image with bounding boxes via OpenCV."""
    if isinstance(image, PIL.Image.Image):
        # noinspection PyTypeChecker
        image = np.asarray(image)
    if not isinstance(image, np.ndarray):
        raise AssertionError("expected input image to be numpy array")
    if not isinstance(rects, list) or not all([isinstance(r, (tuple, list)) and len(r) == 4 for r in rects]):
        raise AssertionError("expected input rectangles to be list of 4-elem tuples/lists (x,y,w,h)")
    if labels is not None and (not isinstance(labels, list) or len(labels) != len(rects)):
        raise AssertionError("bad labels list (check type/length)")
    if confidences is not None and (not isinstance(confidences, list) or len(confidences) != len(confidences)):
        raise AssertionError("bad confidences list (check type/length)")
    display_image = np.copy(image)
    if labels is None and confidences is None:
        # draw all bboxes with unique colors (shuffled)
        rects = copy.deepcopy(rects)
        np.random.shuffle(rects)
        for idx, rect in enumerate(rects):
            cv.rectangle(display_image, (rect[0], rect[1]), (rect[0] + rect[2], rect[1] + rect[3]),
                         get_bgr_from_hsl(idx / len(rects) * 360, 1.0, 0.5), thickness)
    else:
        raise NotImplementedError  # TODO
    if win_size is not None:
        display_image = cv.resize(display_image, win_size)
    if show:
        cv.imshow("bboxes", display_image)
    return display_image


def get_label_color_mapping(idx):
    """Returns the PASCAL VOC color triplet for a given label index."""
    # https://gist.github.com/wllhf/a4533e0adebe57e3ed06d4b50c8419ae
    def bitget(byteval, ch):
        return (byteval & (1 << ch)) != 0
    r = g = b = 0
    for j in range(8):
        r = r | (bitget(idx, 0) << 7 - j)
        g = g | (bitget(idx, 1) << 7 - j)
        b = b | (bitget(idx, 2) << 7 - j)
        idx = idx >> 3
    return np.array([r, g, b], dtype=np.uint8)


def apply_color_map(image, colormap, dst=None):
    """Applies a color map to an image of 8-bit color indices; works similarly to cv2.applyColorMap (v3.3.1)."""
    if not isinstance(image, np.ndarray) or image.ndim != 2:
        raise AssertionError("invalid input image")
    if not isinstance(colormap, np.ndarray) or colormap.shape != (256, 1, 3) or colormap.dtype != np.uint8:
        raise AssertionError("invalid color map")
    out_shape = (image.shape[0], image.shape[1], 3)
    if dst is None:
        dst = np.empty(out_shape, dtype=np.uint8)
    elif not isinstance(dst, np.ndarray) or dst.shape != out_shape or dst.dtype != np.uint8:
        raise AssertionError("invalid output image")
    # using np.take might avoid an extra allocation...
    np.copyto(dst, colormap.squeeze()[image.ravel(), :].reshape(out_shape))
    return dst


def stringify_confmat(confmat, class_list, hide_zeroes=False, hide_diagonal=False, hide_threshold=None):
    """Transforms a confusion matrix array obtained in list or numpy format into a printable string."""
    if not isinstance(confmat, np.ndarray) or not isinstance(class_list, list):
        raise AssertionError("invalid inputs")
    column_width = 9
    empty_cell = " " * column_width
    fst_empty_cell = (column_width - 3) // 2 * " " + "t/p" + (column_width - 3) // 2 * " "
    if len(fst_empty_cell) < len(empty_cell):
        fst_empty_cell = " " * (len(empty_cell) - len(fst_empty_cell)) + fst_empty_cell
    res = "\t" + fst_empty_cell + " "
    for label in class_list:
        res += ("%{0}s".format(column_width) % clipstr(label, column_width)) + " "
    res += ("%{0}s".format(column_width) % "total") + "\n"
    for idx_true, label in enumerate(class_list):
        res += ("\t%{0}s".format(column_width) % clipstr(label, column_width)) + " "
        for idx_pred, _ in enumerate(class_list):
            cell = "%{0}d".format(column_width) % int(confmat[idx_true, idx_pred])
            if hide_zeroes:
                cell = cell if int(confmat[idx_true, idx_pred]) != 0 else empty_cell
            if hide_diagonal:
                cell = cell if idx_true != idx_pred else empty_cell
            if hide_threshold:
                cell = cell if confmat[idx_true, idx_pred] > hide_threshold else empty_cell
            res += cell + " "
        res += ("%{0}d".format(column_width) % int(confmat[idx_true, :].sum())) + "\n"
    res += ("\t%{0}s".format(column_width) % "total") + " "
    for idx_pred, _ in enumerate(class_list):
        res += ("%{0}d".format(column_width) % int(confmat[:, idx_pred].sum())) + " "
    res += ("%{0}d".format(column_width) % int(confmat.sum())) + "\n"
    return res


def fig2array(fig):
    """Transforms a pyplot figure into a numpy-compatible RGB array."""
    fig.canvas.draw()
    w, h = fig.canvas.get_width_height()
    buf = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8)
    buf.shape = (w, h, 3)
    return buf


def get_glob_paths(input_glob_pattern, can_be_dir=False):
    """Parse a wildcard-compatible file name pattern for valid file paths."""
    glob_file_paths = glob.glob(input_glob_pattern)
    if not glob_file_paths:
        raise AssertionError("invalid input glob pattern '%s' (no matches found)" % input_glob_pattern)
    for file_path in glob_file_paths:
        if not os.path.isfile(file_path) and not (can_be_dir and os.path.isdir(file_path)):
            raise AssertionError("invalid input file at globed path '%s'" % file_path)
    return glob_file_paths


def get_file_paths(input_path, data_root, allow_glob=False, can_be_dir=False):
    """Parse a wildcard-compatible file name pattern at a given root level for valid file paths."""
    if os.path.isabs(input_path):
        if '*' in input_path and allow_glob:
            return get_glob_paths(input_path)
        elif not os.path.isfile(input_path) and not (can_be_dir and os.path.isdir(input_path)):
            raise AssertionError("invalid input file at absolute path '%s'" % input_path)
    else:
        if not os.path.isdir(data_root):
            raise AssertionError("invalid dataset root directory at '%s'" % data_root)
        input_path = os.path.join(data_root, input_path)
        if '*' in input_path and allow_glob:
            return get_glob_paths(input_path)
        elif not os.path.isfile(input_path) and not (can_be_dir and os.path.isdir(input_path)):
            raise AssertionError("invalid input file at path '%s'" % input_path)
    return [input_path]
